from datetime import timedelta
import pandas as pd


def twdb_dot(df_row, drop_dcp_metadata=True):
    """Parser for twdb DOT dataloggers."""
    return _twdb_stevens_or_dot(
        df_row, reverse=False, drop_dcp_metadata=drop_dcp_metadata
    )


def twdb_fts(df_row, drop_dcp_metadata=True):
    """Parser for twdb fts dataloggers

    format examples:
    C510D20018036133614G39-0NN170WXW00097  :WL 31 #60 -72.91 -72.89 -72.89 -72.89 -72.91 -72.92 -72.93 -72.96 -72.99 -72.97 -72.95 -72.95
    """

    message = df_row["dcp_message"].lower()
    message_timestamp = df_row["message_timestamp_utc"]

    lines = message.split(":")[1]
    water_levels = [field.strip("+- ") for field in lines.split()[3:]]
    df = _twdb_assemble_dataframe(message_timestamp, None, water_levels, reverse=False)
    return df


def twdb_stevens(df_row, drop_dcp_metadata=True):
    """Parser for twdb stevens dataloggers."""
    return _twdb_stevens_or_dot(
        df_row, reverse=True, drop_dcp_metadata=drop_dcp_metadata
    )


def twdb_sutron(df_row, drop_dcp_metadata=True):
    """Parser for twdb sutron dataloggers.
    Data is transmitted every 12 hours and each message contains 12 water level measurements on the hour
    for the previous 12 hours and one battery voltage measurement for the current hour

    format examples:

    '":ott 60 #60 -190.56 -190.66 -190.69 -190.71 -190.74 -190.73 -190.71 -190.71 -190.71 -190.71 -190.72 -190.72 :BL 13.05  '
    '":SENSE01 60 #60 -82.19 -82.19 -82.18 -82.19 -82.19 -82.22 -82.24 -82.26 -82.27 -82.28 -82.28 -82.26 :BL 12.41  '
    '":OTT 703 60 #60 -231.47 -231.45 -231.44 -231.45 -231.47 -231.50 -231.51 -231.55 -231.56 -231.57 -231.55 -231.53 :6910704 60 #60 -261.85 -261.83 -261.81 -261.80 -261.81 -261.83 -261.85 -261.87 -261.89 -261.88 -261.86 -261.83 :BL 13.21'
    '":Sense01 10 #10 -44.70 -44.68 -44.66 -44.65 -44.63 -44.61 -44.60 -44.57 -44.56 -44.54 -44.52 -44.50 :BL 13.29'
    '"\r\n-101.11 \r\n-101.10 \r\n-101.09 \r\n-101.09 \r\n-101.08 \r\n-101.08 \r\n-101.08 \r\n-101.10 \r\n-101.11 \r\n-101.09 \r\n-101.09 \r\n-101.08'
    '"\r\n// \r\n// \r\n// \r\n// \r\n// \r\n-199.88 \r\n-199.92 \r\n-199.96 \r\n-199.98 \r\n-200.05 \r\n-200.09 \r\n-200.15'
    '":Sense01 60 #60 M M M M M M M M M M M M :BL 12.65'
    """
    message = df_row["dcp_message"].lower()
    message_timestamp = df_row["message_timestamp_utc"]
    lines = message.strip('":').split(":")
    if len(lines) == 1:
        water_levels = [field.strip("+- ") for field in lines[0].split()]
        df = _twdb_assemble_dataframe(
            message_timestamp, None, water_levels, reverse=False
        )
    else:
        data = []
        battery_voltage = lines[-1].split("bl")[-1].strip()
        for line in lines[:-1]:
            channel = line[:7]
            split = line[7:].split()
            water_levels = [field.strip('+-" ') for field in split[2:]]
            df = _twdb_assemble_dataframe(
                message_timestamp, battery_voltage, water_levels, reverse=False
            )
            df["channel"] = channel
            data.append(df)
        df = pd.concat(data)

    if not drop_dcp_metadata:
        for col in df_row.index:
            df[col] = df_row[col]

    return df


def twdb_texuni(dataframe, drop_dcp_metadata=True):
    """Parser for twdb texuni dataloggers.
    Data is transmitted every 12 hours and each message contains 12 water level measurements on the hour
    for the previous 12 hours

    format examples:
    '"\r\n+0.000,-245.3,\r\n+0.000,-245.3,\r\n+0.000,-245.3,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.4,\r\n+0.000,-245.5,\r\n+0.000,-245.5,\r\n+0.000,-245.6,\r\n+0.000,-245.6,\r\n+0.000,-245.6,\r\n+0.000,-245.6,\r\n+0.000,-245.6,\r\n+0.000,-245.6,\r\n+412.0,+2013.,+307.0,+1300.,+12.75,+0.000,-245.4,-245.3,-245.6,+29.55,'
    ' \r\n+0.000,-109.8,\r\n+0.000,-109.8,\r\n+0.000,-109.8,\r\n+0.000,-109.8,\r\n+0.000,-109.8,\r\n+0.000,-109.9,\r\n+0.000,-109.9,\r\n+0.000,-109.9,\r\n+0.000,-109.9,\r\n+0.000,-109.9,\r\n+0.000,-110.0,\r\n+0.000,-110.0,\r\n+0.000,-109.9,\r\n+0.000,-109.9,\r\n+0.000,-109.9,\r\n+0.000,-109.9,\r\n+0.000,-110.0,\r\n+0.000,-110.0,\r\n+0.000,-110.0,\r\n+0.000,-110.1,\r\n+0.000,-110.1,\r\n+0.000,-110.1,\r\n+0.000,-110.1,\r\n+0.000,-110.1,\r\n+340.0,+2013.,+307.0,+1400.,+12.07,+0.000,-109.9,-109.8,-110.1,+30.57,'
    """

    message = dataframe["dcp_message"]
    message_timestamp = dataframe["message_timestamp_utc"]
    water_levels = [
        row.split(",")[1].strip("+- ")
        for row in message.strip('" \r\n').splitlines()[:-1]
    ]

    df = _twdb_assemble_dataframe(message_timestamp, None, water_levels, reverse=True)

    if not drop_dcp_metadata:
        for col in dataframe.index:
            df[col] = dataframe[col]

    return df


def _twdb_assemble_dataframe(
    message_timestamp, battery_voltage, water_levels, reverse=False
):
    data = []
    base_timestamp = message_timestamp.replace(minute=0, second=0, microsecond=0)
    if reverse:
        water_levels.reverse()
    try:
        battery_voltage = float(battery_voltage)
    except:
        battery_voltage = pd.np.nan

    for hrs, water_level in enumerate(water_levels):
        timestamp = base_timestamp - timedelta(hours=hrs)
        try:
            water_level = float(water_level)
        except:
            water_level = pd.np.nan

        if hrs == 0 and battery_voltage:
            data.append([timestamp, battery_voltage, water_level])
        else:
            data.append([timestamp, pd.np.nan, water_level])

    if len(data) > 0:
        df = pd.DataFrame(
            data, columns=["timestamp_utc", "battery_voltage", "water_level"]
        )
        df.index = pd.to_datetime(df["timestamp_utc"])
        del df["timestamp_utc"]
        return df
    else:
        return pd.DataFrame()


def _twdb_stevens_or_dot(df_row, reverse, drop_dcp_metadata=True):
    """Parser for twdb stevens and DOT dataloggers - the only difference being
    that with stevens dataloggers, water level data needs to be reversed to be
    correctly interpretted.

    Data is transmitted every 12 hours and each message contains 12 water level
    measurements on the hour for the previous 12 hours and one battery voltage
    measurement for the current hour

    format examples:

    '"BV:12.5  451.70$ 451.66$ 451.66$ 451.62$ 451.59$ 451.57$ 451.54$ 451.53$ 451.52$ 451.52$ 451.52$ 451.52$ '
    '"BV:12.2  Channel:5 Time:43 +441.48 +443.25 +440.23 +440.67 +441.26 +441.85 +442.66 +443.84 +445.24 +442.15 +442.88 +443.91 '
    '"BV:12.6  Channel:5 Time:28 +304.63 +304.63 +304.63 +304.56 +304.63 +304.63 +304.63 +304.63 +304.63 +304.63 +304.63 +304.71 Channel:6 Time:28 +310.51 +310.66 +310.59 +310.51 +310.51 +310.59 +310.59 +310.51 +310.66 +310.51 +310.66 +310.59 '
    """
    message = df_row["dcp_message"].strip().lower()
    message_timestamp = df_row["message_timestamp_utc"]
    fields = message.strip().split()
    if "bv" in fields[0]:
        battery_voltage = fields[0].split(":")[-1]
        message = " ".join(fields[1:])

    else:
        battery_voltage = None
        message = " ".join(fields)

    fmt = '$+-"\x7f '

    df = []
    if "channel" in message:
        for channel_msg in message.strip("channel:").split("channel:"):
            fields = channel_msg.split()
            msg_channel = fields[0].split(":")[-1]
            msg_time = fields[1].split(":")[-1]
            water_levels = [field.strip(fmt) for field in fields[2:]]
            data = _twdb_assemble_dataframe(
                message_timestamp, battery_voltage, water_levels, reverse=reverse
            )
            data["channel"] = msg_channel
            data["time"] = msg_time
            df.append(data)
    else:
        fields = message.strip(fmt).lstrip().replace(": ", ":").split()
        water_levels = [_parse_value(field.strip(fmt).lstrip()) for field in fields]
        if len(water_levels) and isinstance(water_levels[0], tuple):
            wells = list(set([val[0] for val in water_levels]))
            combined = pd.DataFrame()
            for well in wells:
                values = [
                    val[1]
                    for val in water_levels
                    if isinstance(val, tuple) and val[0] == well
                ]
                data = _twdb_assemble_dataframe(
                    message_timestamp, battery_voltage, values, reverse=reverse
                )
                data.rename(
                    columns={"water_level": "water_level_" + well}, inplace=True
                )
                combined = pd.concat([combined, data], axis=1).T.drop_duplicates().T
            df.append(combined)
        else:
            data = _twdb_assemble_dataframe(
                message_timestamp, battery_voltage, water_levels, reverse=reverse
            )
            df.append(data)

    df = pd.concat(df)

    if not drop_dcp_metadata:
        for col in df_row.index:
            df[col] = df_row[col]

    return df


def _parse_value(water_level_str):
    well_val = water_level_str.split(":")
    if len(water_level_str.split(":")) == 2:
        if well_val[1] == "":
            val = pd.np.nan
        else:
            val = well_val[1].strip("-")
        value_dict = (well_val[0], val)
        return value_dict
    else:
        return water_level_str



from .misc import (
    camel_to_underscore,
    convert_date,
    convert_datetime,
    dict_from_dataframe,
    dir_list,
    download_if_new,
    get_ulmo_dir,
    mkdir_if_doesnt_exist,
    module_with_dependency_errors,
    module_with_deprecation_warnings,
    open_file_for_url,
    parse_fwf,
    raise_dependency_error,
    save_pretty_printed_xml,
    to_bytes,
)

from .raster import (
    extract_from_zip,
    mosaic_and_clip,
    download_tiles,
    generate_raster_uid,
)

try:
    from .pytables import (
        get_default_h5file_path,
        get_or_create_group,
        get_or_create_table,
        open_h5file,
        update_or_append_sortable,
    )
except ImportError:
    get_default_h5file_path = raise_dependency_error
    get_or_create_group = raise_dependency_error
    get_or_create_table = raise_dependency_error
    open_h5file = raise_dependency_error
    update_or_append_sortable = raise_dependency_error



from __future__ import absolute_import

from tsgettoolbox.odo import convert
from tsgettoolbox.odo import odo
from tsgettoolbox.odo import resource

import pandas as pd

from tstoolbox import tsutils

# GHCND
# Global Historical Climatology Network - Daily

# pd.read_csv('ftp://ftp.ncdc.noaa.gov/pub/data/ghcn/daily/all/{0}.dly',
# skiprows=7,
# date_parser=testp,
# index_col=0,
# parse_dates=[[0,1]])


class ghcnd(object):
    def __init__(self, url, **query_params):
        params = {"station": None, "start_date": None, "end_date": None}
        params.update(query_params)

        self.url = url
        self.query_params = params


# Function to make `resource` know about the new ghcnd type.


@resource.register(r"ftp://ftp.ncdc.noaa.gov/pub/data/ghcn/daily/all.*", priority=17)
def resource_ghcnd(uri, **kwargs):
    return ghcnd(uri, **kwargs)


# Function to convert from ghcnd type to pd.DataFrame


@convert.register(pd.DataFrame, ghcnd)
def ghcnd_to_df(data, **kwargs):
    df = pd.read_fwf(
        data.url + "/" + data.query_params["station"] + ".dly",
        widths=[11, 4, 2, 4] + [5, 1, 1, 1] * 31,
    )
    newcols = ["station", "year", "month", "code"]
    for day in list(range(1, 32)):
        for col in ["", "m", "q", "s"]:
            newcols.append("{0}{1:02}".format(col, day))
    df.columns = newcols
    codes = [
        "TMAX",  # Temperature MAX (1/10 degree C)
        "TMIN",  # Temperature MIN (1/10 degree C)
        "PRCP",  # PReCiPitation (tenths of mm)
        "SNOW",  # SNOWfall (mm)
        "SNWD",  # SNoW Depth (mm)
        # Average cloudiness midnight to midnight from 30-second
        # ceilometer data (percent)
        "ACMC",
        # Average cloudiness midnight to midnight from manual observations
        # (percent)
        "ACMH",
        # Average cloudiness sunrise to sunset from 30-second ceilometer
        # data (percent)
        "ACSC",
        # Average cloudiness sunrise to sunset from manual observations
        # (percent)
        "ACSH",
        "AWDR",  # Average daily wind direction (degrees)
        "AWND",  # Average daily wind speed (tenths of meters per second)
        # Number of days included in the multiday evaporation total (MDEV)
        "DAEV",
        # Number of days included in the multiday precipiation total
        # (MDPR)
        "DAPR",
        # Number of days included in the multiday snowfall total (MDSF)
        "DASF",
        # Number of days included in the multiday minimum temperature
        # (MDTN)
        "DATN",
        # Number of days included in the multiday maximum temperature
        # (MDTX)
        "DATX",
        # Number of days included in the multiday wind movement (MDWM)
        "DAWM",
        # Number of days with non-zero precipitation included in multiday
        # precipitation total (MDPR)
        "DWPR",
        # Evaporation of water from evaporation pan (tenths of mm)
        "EVAP",
        # Time of fastest mile or fastest 1-minute wind (hours and
        # minutes, i.e., HHMM)
        "FMTM",
        "FRGB",  # Base of frozen ground layer (cm)
        "FRGT",  # Top of frozen ground layer (cm)
        "FRTH",  # Thickness of frozen ground layer (cm)
        "GAHT",  # Difference between river and gauge height (cm)
        # Multiday evaporation total (tenths of mm; use with DAEV)
        "MDEV",
        # Multiday precipitation total (tenths of mm; use with DAPR and
        # DWPR, if available)
        "MDPR",
        "MDSF",  # Multiday snowfall total
        # Multiday minimum temperature (tenths of degrees C; use with
        # DATN)
        "MDTN",
        # Multiday maximum temperature (tenths of degress C; use with
        # DATX)
        "MDTX",
        "MDWM",  # Multiday wind movement (km)
        # Daily minimum temperature of water in an evaporation pan (tenths
        # of degrees C)
        "MNPN",
        # Daily maximum temperature of water in an evaporation pan (tenths
        # of degrees C)
        "MXPN",
        "PGTM",  # Peak gust time (hours and minutes, i.e., HHMM)
        "PSUN",  # Daily percent of possible sunshine (percent)
        # Average temperature (tenths of degrees C) [Note that TAVG from
        # source 'S' corresponds to an average for the period ending at
        # 2400 UTC rather than local midnight]
        "TAVG",
        "THIC",  # Thickness of ice on water (tenths of mm)
        # Temperature at the time of observation (tenths of degrees C)
        "TOBS",
        "TSUN",  # Daily total sunshine (minutes)
        "WDF1",  # Direction of fastest 1-minute wind (degrees)
        "WDF2",  # Direction of fastest 2-minute wind (degrees)
        "WDF5",  # Direction of fastest 5-second wind (degrees)
        "WDFG",  # Direction of peak wind gust (degrees)
        "WDFI",  # Direction of highest instantaneous wind (degrees)
        "WDFM",  # Fastest mile wind direction (degrees)
        "WDMV",  # 24-hour wind movement (km)
        "WESD",  # Water equivalent of snow on the ground (tenths of mm)
        "WESF",  # Water equivalent of snowfall (tenths of mm)
        # Fastest 1-minute wind speed (tenths of meters per second)
        "WSF1",
        # Fastest 2-minute wind speed (tenths of meters per second)
        "WSF2",
        # Fastest 5-second wind speed (tenths of meters per second)
        "WSF5",
        "WSFG",  # Peak gust wind speed (tenths of meters per second)
        # Highest instantaneous wind speed (tenths of meters per second)
        "WSFI",
        "WSFM",  # Fastest mile wind speed (tenths of meters per second)
    ]

    # SN*# = Minimum soil temperature (tenths of degrees C)
    #        where * corresponds to a code
    #        for ground cover and # corresponds to a code for soil
    #        depth.
    #
    #        Ground cover codes include the following:
    #        0 = unknown
    #        1 = grass
    #        2 = fallow
    #        3 = bare ground
    #        4 = brome grass
    #        5 = sod
    #        6 = straw multch
    #        7 = grass muck
    #        8 = bare muck
    #
    #        Depth codes include the following:
    #        1 = 5 cm
    #        2 = 10 cm
    #        3 = 20 cm
    #        4 = 50 cm
    #        5 = 100 cm
    #        6 = 150 cm
    #        7 = 180 cm
    for i in range(9):
        for j in range(1, 8):
            codes.append("SN{0}{1}".format(i, j))

    # SX*# = Maximum soil temperature (tenths of degrees C)
    #        where * corresponds to a code for ground cover
    #        and # corresponds to a code for soil depth.
    #        See SN*# for ground cover and depth codes.
    for i in range(9):
        for j in range(1, 8):
            codes.append("SX{0}{1}".format(i, j))

    # WT** = Weather Type where ** has one of the following values:
    #
    #        01 = Fog, ice fog, or freezing fog (may include heavy fog)
    #        02 = Heavy fog or heaving freezing fog (not always distinquished
    #        from fog)
    #        03 = Thunder
    #        04 = Ice pellets, sleet, snow pellets, or small hail
    #        05 = Hail (may include small hail)
    #        06 = Glaze or rime
    #        07 = Dust, volcanic ash, blowing dust, blowing sand, or blowing
    #        obstruction
    #        08 = Smoke or haze
    #        09 = Blowing or drifting snow
    #        10 = Tornado, waterspout, or funnel cloud
    #        11 = High or damaging winds
    #        12 = Blowing spray
    #        13 = Mist
    #        14 = Drizzle
    #        15 = Freezing drizzle
    #        16 = Rain (may include freezing rain, drizzle, and freezing
    #        drizzle)
    #        17 = Freezing rain
    #        18 = Snow, snow pellets, snow grains, or ice crystals
    #        19 = Unknown source of precipitation
    #        21 = Ground fog
    #        22 = Ice fog or freezing fog
    codes.extend(["WT{0:02}".format(i) for i in range(1, 23)])

    # WV** = Weather in the Vicinity where ** has one of the following values:
    #        01 = Fog, ice fog, or freezing fog (may include heavy fog)
    #        03 = Thunder
    #        07 = Ash, dust, sand, or other blowing obstruction
    #        18 = Snow or ice crystals
    #        20 = Rain or snow shower
    codes.extend(["WV{0:02}".format(i) for i in [1, 3, 7, 18, 20]])

    for code in codes:
        tmpdf = df.ix[df["code"] == code, :]
        if len(tmpdf) == 0:
            continue
        tmpdf.set_index(["year", "month"], inplace=True)
        tmpdf = tmpdf.ix[:, list(range(2, 126, 4))].stack()
        tmpdf.index = (
            tmpdf.index.get_level_values(0).astype(str).values
            + "-"
            + tmpdf.index.get_level_values(1).astype(str).values
            + "-"
            + tmpdf.index.get_level_values(2).astype(str).values
        )

        # Get rid of bad dates, for example April 31.
        tmpdf.index = pd.to_datetime(tmpdf.index, errors="coerce")
        tmpdf = tmpdf[pd.notnull(tmpdf.index)]

        tmpdf = pd.DataFrame(tmpdf)
        tmpdf.columns = [code]
        tmpdf = tmpdf.ix[
            tsutils.parsedate(
                data.query_params["start_date"], strftime="%Y-%m-%d"
            ) : tsutils.parsedate(data.query_params["end_date"], strftime="%Y-%m-%d"),
            :,
        ]
        try:
            ndf = ndf.join(tmpdf)
        except NameError:
            ndf = tmpdf

    ndf.index.name = "Datetime"
    ndf.replace(to_replace=[-9999], value=[None], inplace=True)
    return ndf


if __name__ == "__main__":
    r = resource(
        r"ftp://ftp.ncdc.noaa.gov/pub/data/ghcn/daily/all",
        station="ASN00075020",
        start_date="2000-01-01",
        end_date="2001-01-01",
    )

    as_df = odo(r, pd.DataFrame)
    print("ghcnd")
    print(as_df)

    r = resource(
        r"ftp://ftp.ncdc.noaa.gov/pub/data/ghcn/daily/all",
        station="ASN00075020",
        start_date="10 years ago",
        end_date="9 years ago",
    )

    as_df = odo(r, pd.DataFrame)
    print("ghcnd")
    print(as_df)



from tsgettoolbox.odo import odo, resource
import pandas as pd
import mando
from tabulate import tabulate as tb

try:
    from mando.rst_text_formatter import RSTHelpFormatter as HelpFormatter
except ImportError:
    from argparse import RawTextHelpFormatter as HelpFormatter

from tstoolbox import tsutils

from tsgettoolbox.services import ldas as placeholder

new_units_table = [
    [
        "{0}\n{1}".format(i, placeholder._UNITS_MAP[i][0]),
        "{0}".format(placeholder._UNITS_MAP[i][1]),
    ]
    for i in placeholder._UNITS_MAP
]

units_table = tb(
    new_units_table,
    tablefmt="grid",
    headers=['LDAS "variable" string Description', "Units"],
)

units_table = "\n".join(["        {0}".format(i) for i in units_table.split("\n")])


@mando.command("ldas", formatter_class=HelpFormatter, doctype="numpy")
@tsutils.doc({"units_table": units_table})
def ldas_cli(
    lat=None,
    lon=None,
    xindex=None,
    yindex=None,
    variable=None,
    startDate=None,
    endDate=None,
):
    r"""Download data from LDAS.

    Available data sources are:

    +-------------------------------+----------+---------------+---------+
    | Description/Name              | Interval | Spatial       | Version |
    +===============================+==========+===============+=========+
    | NLDAS Primary Forcing Data    | 1 hour   | 0.125 x 0.125 | V002    |
    | NLDAS_FORA0125_H              |          | degree        |         |
    +-------------------------------+----------+---------------+---------+
    | NLDAS Noah Land Surface Model | 1 hour   | 0.125 x 0.125 | V002    |
    | NLDAS_NOAH0125_H              |          | degree        |         |
    +-------------------------------+----------+---------------+---------+
    | GLDAS Noah Land Surface Model | 3 hour   | 0.25 x 0.25   | V2.1    |
    | GLDAS_NOAH025_3H              |          | degree        |         |
    +-------------------------------+----------+---------------+---------+
    | AMSR-E/Aqua surface           | 1 day    | 25 x 25 km    | V002    |
    | soil moisture                 |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | AMSR-E/Aqua root zone         | 1 day    | 25 x 25 km    | V001    |
    | soil moisture                 |          |               |         |
    | LPRM_AMSRE_D_RZSM3            |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | AMSR2/GCOM-W1 surface         | 1 day    | 25 x 25 km    | V001    |
    | soil moisture                 |          |               |         |
    | LPRM_AMSR2_A_SOILM3           |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | AMSR2/GCOM-W1 surface         | 1 day    | 25 x 25 km    | V001    |
    | soil moisture                 |          |               |         |
    | LPRM_AMSR2_D_SOILM3           |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | AMSR2/GCOM-W1 surface         | 1 day    | 10 x 10 km    | V001    |
    | soil moisture                 |          |               |         |
    | LPRM_AMSR2_DS_A_SOILM3        |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | ASMR2/GCOM-W1 surface         | 1 day    | 10 x 10 km    | V001    |
    | soil moisture                 |          |               |         |
    | LPRM_AMSR2_DS_D_SOILM3        |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | TMI/TRMM surface              | 1 day    | 25 x 25 km    | V001    |
    | soil moisture                 |          |               |         |
    | LPRM_TMI_NT_SOILM3            |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | TMI/TRMM surface              | 1 day    | 25 x 25 km    | V001    |
    | soil moisture                 |          |               |         |
    | LPRM_TMI_DY_SOILM3            |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | TRMM (TMPA) Rainfall Estimate | 3 hour   | 0.25 x 0.25   | V7      |
    | TRMM_3B42                     |          | degree        |         |
    +-------------------------------+----------+---------------+---------+
    | Smerge-Noah-CCI root zone     | day      | 0.125 x 0.125 | V2.0    |
    | soil moisture 0-40 cm         |          | degree        |         |
    | SMERGE_RZSM0_40CM             |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | Groundwater and Soil Moisture | 7 day    | 0.125 x 0.125 | V2.0    |
    | Conditions from GRACE         |          | degree        |         |
    | Data Assimilation             |          |               |         |
    | GRACEDADM_CLSM0125US_7D       |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | MERRA-2 2D, Instantaneous,    | 1 hour   |               | V5.12.4 |
    | Land Surface Forcings         |          |               |         |
    | M2I1NXLFO                     |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | MERRA-2 2D, Time-averaged,    | 1 hour   |               | V5.12.4 |
    | Surface Flux Diagnostics      |          |               |         |
    | M2T1NXFLX                     |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | MERRA-2 2D, Time-averaged,    | 1 hour   |               | V5.12.4 |
    | Land Surface Forcings         |          |               |         |
    | M2T1NXLFO                     |          |               |         |
    +-------------------------------+----------+---------------+---------+
    | MERRA 2D Incremental          | 1 hour   |               | V5.2.0  |
    | Analysis Update               |          |               |         |
    | MST1NXMLD                     |          |               |         |
    +-------------------------------+----------+---------------+---------+

    The time zone is always UTC.

    Parameters
    ----------
    lat :  float
        Should use 'lat' and 'lon' to specify location.

        Latitude (required): Enter single geographic point by latitude.::

            Example: --lat=43.1

        If known, 'xindex' and 'yindex' can be used for the NLDAS grid only.
    lon : float
        Should use 'lat' and 'lon' to specify location.

        Longitude (required): Enter single geographic point by longitude::

            Example: --lon=-85.3

        If known, 'xindex' and 'yindex' can be used for the NLDAS grid only.
    xindex : int
        It `lat` or `lon` is None, then will try `xindex` and `yindex`.

        Enter the x index of the NLDAS grid.::

            Example: --xindex=301

    yindex : int
        It `lat` or `lon` is None, then will try `xindex` and `yindex`.

        Enter the y index of the NLDAS grid.::

            Example: --yindex=80

    variable : str
        Use the variable codes from the following table:

{units_table}

    startDate : str
        The start date of the time series.::

            Example: --startDate=2001-01-01T05

        If startDate and endDate are None, returns the entire series.

    endDate : str
        The end date of the time series.::

            Example: --endDate=2002-01-05T05

        If startDate and endDate are None, returns the entire series.
    """
    tsutils._printiso(
        ldas(
            lat=lat,
            lon=lon,
            xindex=xindex,
            yindex=yindex,
            variable=variable,
            startDate=startDate,
            endDate=endDate,
        )
    )


@tsutils.validator(
    lat=[float, ["range", [-90.0, 90.0]], 1],
    lon=[float, ["range", [-180.0, 180.0]], 1],
    xindex=[int, ["range", [0, None]], 1],
    yindex=[int, ["range", [0, None]], 1],
    variable=[str, ["pass", []], None],
    startDate=[tsutils.parsedate, ["pass", []], 1],
    endDate=[tsutils.parsedate, ["pass", []], 1],
)
def ldas(
    lat=None,
    lon=None,
    xindex=None,
    yindex=None,
    variable=None,
    startDate=None,
    endDate=None,
):
    r"""Download data from NLDAS or GLDAS."""
    project = variable.split(":")[0]
    if lat is not None and lon is not None:
        location = "GEOM:POINT({0}, {1})".format(lon, lat)
    elif project == "NLDAS" and xindex is not None and yindex is not None:
        location = "{0}:X{1:03d}-Y{2:03d}".format(project, xindex, yindex)
    else:
        raise ValueError(
            tsutils.error_wrapper(
                """
There is a problem specifying the location.

Both `lat` and `lon` need to be specified where you have "lat={lat}" and
"lon={lon}".

Only for the NLDAS grid can you use `xindex` and `yindex` to specify the
location.  You have the grid "{project}" and "xindex={xindex}" and
"yindex={yindex}".
""".format(
                    **locals()
                )
            )
        )

    ndf = pd.DataFrame()
    for cnt, var in enumerate(tsutils.make_list(variable)):
        r = resource(
            r"https://hydro1.gesdisc.eosdis.nasa.gov/daac-bin/access/timeseries.cgi",
            variable=var,
            location=location,
            startDate=startDate,
            endDate=endDate,
        )
        ndf = ndf.join(odo(r, pd.DataFrame), how="outer", rsuffix="_{0}".format(cnt))

    return ndf


ldas.__doc__ = ldas_cli.__doc__



from __future__ import absolute_import, division, print_function

import os
from tsgettoolbox.odo.resource import resource
from tsgettoolbox.odo.utils import raises


def test_raises_not_implemented_error():
    assert raises(
        NotImplementedError, lambda: resource("5sdjkg9yg35420shfg083.3923.925y2560!")
    )



import os
import sys
import shutil
import pytest


@pytest.fixture(scope="session")
def sc():
    pytest.importorskip("pyspark")
    from pyspark import SparkContext

    return SparkContext("local[*]", "odo")


@pytest.yield_fixture(scope="session")
def sqlctx(sc):
    pytest.importorskip("pyspark")
    from tsgettoolbox.odo.backends.sparksql import HiveContext

    try:
        yield HiveContext(sc)
    finally:
        dbpath = "metastore_db"
        logpath = "derby.log"
        if os.path.exists(dbpath):
            assert os.path.isdir(dbpath)
            shutil.rmtree(dbpath)
        if os.path.exists(logpath):
            assert os.path.isfile(logpath)
            os.remove(logpath)



from __future__ import absolute_import, division, print_function

import numpy as np
import datashape as ds
import pytest

tb = pytest.importorskip("tables")

from tsgettoolbox.odo import into, odo
from tsgettoolbox.odo.utils import tmpfile
from tsgettoolbox.odo.backends.pytables import PyTables, discover
import os


try:
    f = tb.open_file("import-tables-test.hdf5", mode="w")
    f.close()
    if os.path.exists("import-tables-test.hdf5"):
        os.remove("import-tables-test.hdf5")
except tb.exceptions.HDF5ExtError as e:
    pytest.skip("Cannot write file, error:\n%s" % e)


x = np.array(
    [
        (1, "Alice", 100),
        (2, "Bob", -200),
        (3, "Charlie", 300),
        (4, "Denis", 400),
        (5, "Edith", -500),
    ],
    dtype=[("id", "<i8"), ("name", "S7"), ("amount", "<i8")],
)


@pytest.yield_fixture
def tbfile():
    with tmpfile(".h5") as filename:
        f = tb.open_file(filename, mode="w")
        d = f.create_table("/", "title", x)
        d.close()
        f.close()
        yield filename


now = np.datetime64("now").astype("datetime64[us]")
raw_dt_data = [
    (1, "Alice", 100, now),
    (2, "Bob", -200, now),
    (3, "Charlie", 300, now),
    (4, "Denis", 400, now),
    (5, "Edith", -500, now),
]


dt_data = np.array(
    raw_dt_data,
    dtype=np.dtype(
        [("id", "i8"), ("name", "S7"), ("amount", "f8"), ("date", "M8[ms]")]
    ),
)


@pytest.yield_fixture
def dt_tb():
    class Desc(tb.IsDescription):
        id = tb.Int64Col(pos=0)
        name = tb.StringCol(itemsize=7, pos=1)
        amount = tb.Float64Col(pos=2)
        date = tb.Time64Col(pos=3)

    non_date_types = list(zip(["id", "name", "amount"], ["i8", "S7", "f8"]))

    # has to be in microseconds as per pytables spec
    dtype = np.dtype(non_date_types + [("date", "M8[us]")])
    rec = dt_data.astype(dtype)

    # also has to be a floating point number
    dtype = np.dtype(non_date_types + [("date", "f8")])
    rec = rec.astype(dtype)
    rec["date"] /= 1e6
    with tmpfile(".h5") as filename:
        f = tb.open_file(filename, mode="w")
        d = f.create_table("/", "dt", description=Desc)
        d.append(rec)
        d.close()
        f.close()
        yield filename


class TestPyTablesLight(object):
    def test_read(self, tbfile):
        t = PyTables(path=tbfile, datapath="/title")
        shape = t.shape
        t._v_file.close()
        assert shape == (5,)

    def test_write_no_dshape(self, tbfile):
        with pytest.raises(ValueError):
            PyTables(path=tbfile, datapath="/write_this")

    def test_write_with_dshape(self, tbfile):
        f = tb.open_file(tbfile, mode="a")
        try:
            assert "/write_this" not in f
        finally:
            f.close()
            del f

        # create our table
        dshape = '{id: int, name: string[7, "ascii"], amount: float32}'
        t = PyTables(path=tbfile, datapath="/write_this", dshape=dshape)
        shape = t.shape
        filename = t._v_file.filename
        t._v_file.close()

        assert filename == tbfile
        assert shape == (0,)

    @pytest.mark.xfail(reason="Poor datetime support")
    def test_table_into_ndarray(self, dt_tb):
        t = PyTables(dt_tb, "/dt")
        res = into(np.ndarray, t)
        try:
            for k in res.dtype.fields:
                lhs, rhs = res[k], dt_data[k]
                if issubclass(np.datetime64, lhs.dtype.type) and issubclass(
                    np.datetime64, rhs.dtype.type
                ):
                    lhs, rhs = lhs.astype("M8[us]"), rhs.astype("M8[us]")
                assert np.array_equal(lhs, rhs)
        finally:
            t._v_file.close()

    def test_ndarray_into_table(self, dt_tb):
        dtype = ds.from_numpy(dt_data.shape, dt_data.dtype)
        t = PyTables(dt_tb, "/out", dtype)
        try:
            res = into(np.ndarray, into(t, dt_data, filename=dt_tb, datapath="/out"))
            for k in res.dtype.fields:
                lhs, rhs = res[k], dt_data[k]
                if issubclass(np.datetime64, lhs.dtype.type) and issubclass(
                    np.datetime64, rhs.dtype.type
                ):
                    lhs, rhs = lhs.astype("M8[us]"), rhs.astype("M8[us]")
                assert np.array_equal(lhs, rhs)
        finally:
            t._v_file.close()

    @pytest.mark.xfail(reason="Poor datetime support")
    def test_datetime_discovery(self, dt_tb):
        t = PyTables(dt_tb, "/dt")
        lhs, rhs = map(discover, (t, dt_data))
        t._v_file.close()
        assert lhs == rhs

    def test_node_discover(self, dt_tb):
        root = PyTables(dt_tb, "/")
        result = discover(root)
        expected = ds.dshape(
            """{dt: 5 * {id: int64,
                                          name: string[7, "A"],
                                          amount: float64,
                                          date: float64}}"""
        )
        assert result == expected.measure
        root._v_file.close()

    def test_no_extra_files_around(self, dt_tb):
        """ check the context manager auto-closes the resources """
        assert not len(tb.file._open_files)


def test_pytables_to_csv():
    ndim = 2
    with tmpfile(".h5") as fn:
        h5file = tb.open_file(fn, mode="w", title="Test Array")
        h5file.create_array("/", "test", np.zeros((ndim, ndim), dtype=float))
        h5file.close()
        with tmpfile("csv") as csv:
            t = odo("pytables://%s::/test" % fn, csv)
            assert odo(t, list) == [(0.0, 0.0), (0.0, 0.0)]



from __future__ import absolute_import, division, print_function

import pytest

sas7bdat = pytest.importorskip("sas7bdat")
pytest.importorskip("odo.backends.sas")
import os
import pandas as pd
from collections import Iterator
from sas7bdat import SAS7BDAT

from tsgettoolbox.odo.backends.sas import discover, sas_to_iterator
from tsgettoolbox.odo.utils import tmpfile, into_path
from tsgettoolbox.odo import append, convert, resource, dshape


test_path = into_path("backends", "tests", "airline.sas7bdat")
sasfile = SAS7BDAT(test_path)


columns = (
    "DATE",
    "AIR",
    "mon1",
    "mon2",
    "mon3",
    "mon4",
    "mon5",
    "mon6",
    "mon7",
    "mon8",
    "mon9",
    "mon10",
    "mon11",
    "mon12",
    "t",
    "Lair",
)

ds = dshape(
    """var * {DATE: date, AIR: float64, mon1: float64, mon2: float64,
                      mon3: float64, mon4: float64, mon5: float64,
                      mon6: float64, mon7: float64, mon8: float64,
                      mon9: float64, mon10: float64, mon11: float64,
                      mon12: float64, t: float64, Lair: float64}"""
)


def test_resource_sas7bdat():
    assert isinstance(resource(test_path), SAS7BDAT)


def test_discover_sas():
    assert discover(sasfile) == ds


def test_convert_sas_to_dataframe():
    df = convert(pd.DataFrame, sasfile)
    assert isinstance(df, pd.DataFrame)

    # pandas doesn't support date
    expected = str(ds.measure).replace("date", "datetime")

    assert str(discover(df).measure).replace("?", "") == expected


def test_convert_sas_to_list():
    out = convert(list, sasfile)
    assert isinstance(out, list)
    assert not any(isinstance(item, str) for item in out[0])  # No header
    assert all(isinstance(ln, list) for ln in out)


def test_convert_sas_to_iterator():
    itr = sas_to_iterator(sasfile)
    assert isinstance(itr, Iterator)


def test_append_sas_to_sqlite_round_trip():
    expected = convert(set, sasfile)

    with tmpfile("db") as fn:
        r = resource("sqlite:///%s::SAS" % fn, dshape=discover(sasfile))
        append(r, sasfile)

        result = convert(set, r)

    assert expected == result



"""
Provide access to data from the USACE `Rivergages`_ web site.

`United States Army Corps of Engineers`_ `Rivergages`_ web site.

.. _United States Army Corps of Engineers: http://www.usace.army.mil/
.. _Rivergages: http://rivergages.mvr.usace.army.mil/WaterControl/new/layout.cfm
"""
import pandas as pd

from tstoolbox import tsutils

from tsgettoolbox.ulmo.usace.rivergages.core import get_station_data
from tsgettoolbox.ulmo.usace.rivergages.core import get_station_parameters
from tsgettoolbox.ulmo.usace.rivergages.core import get_stations

# def get_station_data(station_code, parameter, start=None, end=None,
#         min_value=None, max_value=None):


def ulmo_df(station_code, parameter, start_date=None, end_date=None):
    tstations = get_stations()
    if station_code not in tstations:
        raise ValueError(
            tsutils.error_wrapper(
                """
Station code {0} not in available stations:
{1}
""".format(
                    station_code, tstations.keys
                )
            )
        )

    tparameters = get_station_parameters(station_code)
    if parameter not in tparameters:
        raise ValueError(
            tsutils.error_wrapper(
                """
Parameter code {0} not in available parameters at station {1}:
{2}
""".format(
                    parameter, station_code, tparameters
                )
            )
        )
    df = get_station_data(
        station_code,
        parameter,
        start=pd.to_datetime(start_date),
        end=pd.to_datetime(end_date),
    )
    df = pd.DataFrame.from_dict(df, orient="index")
    df.sort_index(inplace=True)
    df.index.name = "Datetime"
    df.columns = ["{0}_{1}".format(station_code, parameter)]
    return df


if __name__ == "__main__":
    #    import time
    #
    #    r = ulmo_df('blah',
    #                'upperbasin')
    #
    #    print('BIVOI_HL')
    #    print(r)
    #
    r = ulmo_df("BIVO1", "HL", start_date="2015-11-04", end_date="2015-12-05")

    print("BIVOI HL")
    print(r)



from __future__ import absolute_import, division, print_function
import warnings

from tsgettoolbox.odo.core import NetworkDispatcher, path, FailedConversionWarning
from datashape import discover

d = NetworkDispatcher("foo")


@d.register(float, int, cost=1.0)
def f(x, **kwargs):
    return float(x)


@d.register(str, float, cost=1.0)
def g(x, **kwargs):
    return str(x)


def test_basic():
    assert [func for a, b, func, cost in d.path(int, str)] == [f, g]

    assert list(d.path(int, str)) == list(d.path(1, ""))


def test_convert_is_robust_to_failures():
    foo = NetworkDispatcher("foo")

    def badfunc(*args, **kwargs):
        raise NotImplementedError()

    class A(object):
        pass

    class B(object):
        pass

    class C(object):
        pass

    discover.register((A, B, C))(lambda x: "int")
    foo.register(B, A, cost=1.0)(lambda x, **kwargs: 1)
    foo.register(C, B, cost=1.0)(badfunc)
    foo.register(C, A, cost=10.0)(lambda x, **kwargs: 2)

    with warnings.catch_warnings(record=True) as ws:
        warnings.simplefilter("always")
        assert foo(C, A()) == 2

    assert len(ws) == 1
    w = ws[0].message
    assert isinstance(w, FailedConversionWarning)
    assert "B -> C" in str(w)


def test_convert_failure_takes_greedy_path():
    foo = NetworkDispatcher("foo")

    class A(object):
        pass

    class B(object):
        pass

    class C(object):
        pass

    class D(object):
        pass

    discover.register((A, B, C, D))(lambda _: "int")

    foo.register(B, A, cost=1.0)(lambda _, **__: 1)

    @foo.register(D, A, cost=10.0)
    def expensive_edge(*args, **kwargs):
        raise AssertionError(
            "convert should not take this route because it is more expensive"
            " than the initial route and greedy route"
        )

    @foo.register(C, B, cost=1.0)
    def badedge(*args, **kwargs):
        raise NotImplementedError()

    @foo.register(D, C, cost=1.0)
    def impossible_edge(*args, **kwargs):
        raise AssertionError(
            "to get to this edge B->C would need to pass which is impossible"
        )

    greedy_retry_route_selected = [False]

    # this edge is more expensive than the cost of B->C->D so it shouldn't
    # be picked until B->C has been removed
    @foo.register(D, B, cost=3.0)
    def greedy_retry_route(data, **kwargs):
        greedy_retry_route_selected[0] = True
        assert data == 1
        return 2

    with warnings.catch_warnings(record=True) as ws:
        warnings.simplefilter("always")
        assert foo(D, A()) == 2

    assert greedy_retry_route_selected[0], "we did not call the expected edge"

    assert len(ws) == 1
    w = ws[0].message
    assert isinstance(w, FailedConversionWarning)
    assert "B -> C" in str(w)


def test_ooc_behavior():
    foo = NetworkDispatcher("foo")

    class A(object):
        pass

    class B(object):
        pass

    class C(object):
        pass

    discover.register((A, B, C))(lambda x: "int")
    foo.register(B, A, cost=1.0)(lambda x, **kwargs: 1)
    foo.register(C, B, cost=1.0)(lambda x, **kwargs: x / 0)  # note that this errs
    foo.register(C, A, cost=10.0)(lambda x, **kwargs: 2)

    assert [(a, b, cost) for a, b, _, cost in path(foo.graph, A, C)] == [
        (A, B, 1.0),
        (B, C, 1.0),
    ]

    ooc = set([A, C])
    assert [
        (a, b, cost) for a, b, _, cost, in path(foo.graph, A, C, ooc_types=ooc)
    ] == [(A, C, 10.0)]



from __future__ import print_function

import pytest
import sys

pytestmark = pytest.mark.skipif(sys.platform == "win32", reason="Requires Mac or Linux")

boto = pytest.importorskip("boto")

import os
import itertools
import json
from contextlib import contextmanager, closing

import datashape
from datashape import string, float64, int64
from datashape.util.testing import assert_dshape_equal
import pandas as pd
import pandas.util.testing as tm

from tsgettoolbox.odo import into, resource, S3, discover, CSV, drop, append, odo
from tsgettoolbox.odo.backends.aws import get_s3_connection
from tsgettoolbox.odo.utils import tmpfile
from tsgettoolbox.odo.compatibility import urlopen


from boto.exception import S3ResponseError

tips_uri = "s3://nyqpug/tips.csv"

df = pd.DataFrame({"a": list("abc"), "b": [1, 2, 3], "c": [1.0, 2.0, 3.0]})[
    ["a", "b", "c"]
]


js = pd.io.json.loads(pd.io.json.dumps(df, orient="records"))

is_authorized = False
tried = False

with closing(urlopen("http://httpbin.org/ip")) as url:
    public_ip = json.loads(url.read().decode())["origin"]

cidrip = public_ip + "/32"


@pytest.yield_fixture
def tmpcsv():
    with tmpfile(".csv") as fn:
        with open(fn, mode="w") as f:
            df.to_csv(f, index=False)
        yield fn


@contextmanager
def s3_bucket(extension):
    with conn():
        b = "s3://%s/%s%s" % (test_bucket_name, next(_tmps), extension)
        try:
            yield b
        finally:
            drop(resource(b))


@contextmanager
def conn():
    # requires that you have a config file or envars defined for credentials
    # this code makes me hate exceptions
    try:
        conn = get_s3_connection()
    except S3ResponseError:
        pytest.skip("unable to connect to s3")
    else:
        try:
            grants = conn.get_bucket(test_bucket_name).get_acl().acl.grants
        except S3ResponseError:
            pytest.skip("no permission to read on bucket %s" % test_bucket_name)
        else:
            if not any(
                g.permission == "FULL_CONTROL" or g.permission == "READ" for g in grants
            ):
                pytest.skip("no permission to read on bucket %s" % test_bucket_name)
            else:
                yield conn


test_bucket_name = "into-redshift-csvs"

_tmps = ("tmp%d" % i for i in itertools.count())


@pytest.fixture
def s3_encryption_bucket():
    test_bucket = os.getenv("ODO_S3_ENCRYPTION_BUCKET")
    if not test_bucket:
        pytest.skip("No bucket defined that requires server-side encryption")
    return test_bucket


def test_s3_encrypted_upload(s3_encryption_bucket):
    s3_connection = boto.connect_s3()

    df = tm.makeMixedDataFrame()
    with tmpfile(".csv") as fn:
        df.to_csv(fn, index=False)
        s3_uri = "s3://{bucket}/{fn}".format(
            bucket=s3_encryption_bucket, fn=os.path.basename(fn)
        )
        odo(fn, s3_uri, s3=s3_connection, encrypt_key=True)
        result = odo(s3_uri, pd.DataFrame, s3=s3_connection)

    tm.assert_frame_equal(df, result)


def test_s3_encrypted_multipart_upload(s3_encryption_bucket):
    s3_connection = boto.connect_s3()

    df = tm.makeMixedDataFrame()
    with tmpfile(".csv") as fn:
        df.to_csv(fn, index=False)
        s3_uri = "s3://{bucket}/{fn}".format(
            bucket=s3_encryption_bucket, fn=os.path.basename(fn)
        )
        odo(fn, s3_uri, s3=s3_connection, encrypt_key=True, multipart=True)
        result = odo(s3_uri, pd.DataFrame, s3=s3_connection)

    tm.assert_frame_equal(df, result)


def test_s3_resource():
    csv = resource(tips_uri)
    assert isinstance(csv, S3(CSV))


def test_s3_discover():
    csv = resource(tips_uri)
    assert isinstance(discover(csv), datashape.DataShape)


def test_s3_to_local_csv():
    with tmpfile(".csv") as fn:
        csv = into(fn, tips_uri)
        path = os.path.abspath(csv.path)
        assert os.path.exists(path)


def test_csv_to_s3_append():
    df = tm.makeMixedDataFrame()
    with tmpfile(".csv") as fn:
        with s3_bucket(".csv") as b:
            s3 = resource(b)
            df.to_csv(fn, index=False)
            append(s3, CSV(fn))
            result = into(pd.DataFrame, s3)
    tm.assert_frame_equal(df, result)


def test_csv_to_s3_into():
    df = tm.makeMixedDataFrame()
    with tmpfile(".csv") as fn:
        with s3_bucket(".csv") as b:
            df.to_csv(fn, index=False)
            s3 = into(b, CSV(fn))
            result = into(pd.DataFrame, s3)
    tm.assert_frame_equal(df, result)


def test_frame_to_s3_to_frame():
    with s3_bucket(".csv") as b:
        s3_csv = into(b, df)
        result = into(pd.DataFrame, s3_csv)
    tm.assert_frame_equal(result, df)


def test_textfile_to_s3():
    text = "A cow jumped over the moon"
    with tmpfile(".txt") as fn:
        with s3_bucket(".txt") as b:
            with open(fn, mode="w") as f:
                f.write(os.linesep.join(text.split()))
            result = into(b, resource(fn))
    assert discover(result) == datashape.dshape("var * string")


def test_jsonlines_to_s3():
    with tmpfile(".json") as fn:
        with open(fn, mode="w") as f:
            for row in js:
                f.write(pd.io.json.dumps(row))
                f.write(os.linesep)
        with s3_bucket(".json") as b:
            result = into(b, resource(fn))
            assert discover(result) == discover(js)


def test_s3_jsonlines_discover():
    json_dshape = discover(resource("s3://nyqpug/tips.json"))
    names = list(map(str, sorted(json_dshape.measure.names)))
    assert names == ["day", "sex", "size", "smoker", "time", "tip", "total_bill"]
    types = [json_dshape.measure[name] for name in names]
    assert types == [string, string, int64, string, string, float64, float64]


def test_s3_csv_discover():
    result = discover(resource("s3://nyqpug/tips.csv"))
    expected = datashape.dshape(
        """var * {
      total_bill: float64,
      tip: float64,
      sex: ?string,
      smoker: ?string,
      day: ?string,
      time: ?string,
      size: int64
      }"""
    )
    assert_dshape_equal(result, expected)


def test_s3_gz_csv_discover():
    result = discover(S3(CSV)("s3://nyqpug/tips.gz"))
    expected = datashape.dshape(
        """var * {
      total_bill: float64,
      tip: float64,
      sex: ?string,
      smoker: ?string,
      day: ?string,
      time: ?string,
      size: int64
      }"""
    )
    assert_dshape_equal(result, expected)


def test_s3_to_sqlite():
    with tmpfile(".db") as fn:
        tb = into(
            "sqlite:///%s::tips" % fn, tips_uri, dshape=discover(resource(tips_uri))
        )
        lhs = into(list, tb)
        assert lhs == into(list, tips_uri)


def test_csv_to_s3__using_multipart_upload():
    df = pd.DataFrame({"a": ["*" * 5 * 1024 ** 2]})
    with tmpfile(".csv") as fn:
        with s3_bucket(".csv") as b:
            df.to_csv(fn, index=False)
            s3 = into(b, CSV(fn), multipart=True)
            result = into(pd.DataFrame, s3)
    tm.assert_frame_equal(df, result)


@pytest.mark.parametrize(
    ["prefix", "suffix"],
    [pytest.mark.xfail(("xa", ""), raises=NotImplementedError), ("za", ".csv")],
)
def test_chunks_of_s3(prefix, suffix):
    uri = "s3://nyqpug/{}*{}".format(prefix, suffix)
    result = resource(uri)
    assert len(result.data) == 2
    expected = odo(tips_uri, pd.DataFrame)
    tm.assert_frame_equal(odo(result, pd.DataFrame), expected)



from __future__ import print_function
from __future__ import absolute_import

import logging
import os
from builtins import object
from io import BytesIO

from tsgettoolbox.odo import convert
from tsgettoolbox.odo import odo
from tsgettoolbox.odo import resource

import pandas as pd

import requests

from tstoolbox import tsutils

_UNITS_MAP = {
    "NLDAS:NLDAS_FORA0125_H.002:APCPsfc": ["Precipitation hourly total", "kg/m2"],
    "NLDAS:NLDAS_FORA0125_H.002:DLWRFsfc": [
        "Surface DW longwave radiation flux",
        "W/m2",
    ],
    "NLDAS:NLDAS_FORA0125_H.002:DSWRFsfc": [
        "Surface DW shortwave radiation flux",
        "W/m2",
    ],
    "NLDAS:NLDAS_FORA0125_H.002:PEVAPsfc": ["Potential evaporation", "kg/m2"],
    "NLDAS:NLDAS_FORA0125_H.002:SPFH2m": [
        "2-m above ground specific humidity",
        "kg/kg",
    ],
    "NLDAS:NLDAS_FORA0125_H.002:TMP2m": ["2-m above ground temperature", "K"],
    "NLDAS:NLDAS_FORA0125_H.002:UGRD10m": ["10-m above ground zonal wind", "m/s"],
    "NLDAS:NLDAS_FORA0125_H.002:VGRD10m": ["10-m above ground meridional wind", "m/s"],
    "NLDAS:NLDAS_NOAH0125_H.002:EVPsfc": ["Total evapotranspiration", "kg/m2"],
    "NLDAS:NLDAS_NOAH0125_H.002:GFLUXsfc": ["Ground heat flux", "W/m2"],
    "NLDAS:NLDAS_NOAH0125_H.002:LHTFLsfc": ["Latent heat flux", "W/m2"],
    "NLDAS:NLDAS_NOAH0125_H.002:SHTFLsfc": ["Sensible heat flux", "W/m2"],
    "NLDAS:NLDAS_NOAH0125_H.002:SSRUNsfc": [
        "Surface runoff (non-infiltrating)",
        "kg/m2",
    ],
    "NLDAS:NLDAS_NOAH0125_H.002:BGRUNsfc": ["Subsurface runoff (baseflow)", "kg/m2"],
    "NLDAS:NLDAS_NOAH0125_H.002:SOILM0-10cm": [
        "0-10 cm soil moisture content",
        "kg/m2",
    ],
    "NLDAS:NLDAS_NOAH0125_H.002:SOILM0-100cm": [
        "0-100 cm soil moisture content",
        "kg/m2",
    ],
    "NLDAS:NLDAS_NOAH0125_H.002:SOILM0-200cm": [
        "0-200 cm soil moisture content",
        "kg/m2",
    ],
    "NLDAS:NLDAS_NOAH0125_H.002:SOILM10-40cm": [
        "10-40 cm soil moisture content",
        "kg/m2",
    ],
    "NLDAS:NLDAS_NOAH0125_H.002:SOILM40-100cm": [
        "40-100 cm soil moisture content",
        "kg/m2",
    ],
    "NLDAS:NLDAS_NOAH0125_H.002:SOILM100-200cm": [
        "100-200 cm soil moisture content",
        "kg/m2",
    ],
    "NLDAS:NLDAS_NOAH0125_H.002:TSOIL0-10cm": ["0-10 cm soil temperature", "K"],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:AvgSurfT_inst": [
        "Average surface skin temperature",
        "K",
    ],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Evap_tavg": ["Evapotranspiration", "kg/m2/s"],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Psurf_f_inst": ["Surface air pressure", "Pa"],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Qair_f_inst": ["Specific humidity", "kg/kg"],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Qs_acc": ["Storm surface runoff", "kg/m2"],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Qsb_acc": ["Baseflow-groundwater runoff", "kg/m2"],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Rainf_f_tavg": [
        "Total precipitation rate",
        "kg/m2/s",
    ],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Rainf_tavg": ["Rain precipitation rate", "kg/m2/s"],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Snowf_tavg": ["Snow precipitation rate", "kg/m2/s"],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:SoilMoi0_10cm_inst": [
        "Soil moisture content (0-10 cm underground)",
        "kg/m2",
    ],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:SoilMoi10_40cm_inst": [
        "Soil moisture content (10-40 cm underground)",
        "kg/m2",
    ],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:SoilMoi40_100cm_inst": [
        "Soil moisture content (40-100 cm underground)",
        "kg/m2",
    ],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:SoilTMP0_10cm_inst": [
        "Soil temperature (0-10 cm underground)",
        "K",
    ],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Tair_f_inst": ["Near surface air temperature", "K"],
    "GLDAS2:GLDAS_NOAH025_3H_v2.1:Wind_f_inst": ["Near surface wind speed", "m/s"],
    "LPRM:LPRM_AMSRE_D_SOILM3.002:opt_depth_c": [
        "Optical depth from LPRM AMSRE C-band descending",
        "unitless",
    ],
    "LPRM:LPRM_AMSRE_D_SOILM3.002:opt_depth_x": [
        "Optical depth from LPRM AMSRE X-band descending",
        "unitless",
    ],
    "LPRM:LPRM_AMSRE_D_SOILM3.002:sm_c_error": [
        "Soil moisture uncertainty of LPRM AMSRE C-band",
        "percent",
    ],
    "LPRM:LPRM_AMSRE_D_SOILM3.002:sm_x_error": [
        "Soil moisture uncertainty of LPRM AMSRE X-band",
        "percent",
    ],
    "LPRM:LPRM_AMSRE_D_SOILM3.002:soil_moisture_c": [
        "Soil moisture, volumetric, from LPRM AMSRE C-band descending",
        "percent",
    ],
    "LPRM:LPRM_AMSRE_D_SOILM3.002:soil_moisture_x": [
        "Soil moisture, volumetric, from LPRM AMSRE X-band descending",
        "percent",
    ],
    "LPRM:LPRM_AMSRE_D_SOILM3.002:Ts": [
        "Skin temperature (2mm) from LPRM AMSRE descending",
        "K",
    ],
    "LPRM:LPRM_AMSRE_A_SOILM3.002:opt_depth_c": [
        "Optical depth from LPRM AMSRE C-band descending",
        "unitless",
    ],
    "LPRM:LPRM_AMSRE_A_SOILM3.002:opt_depth_x": [
        "Optical depth from LPRM AMSRE X-band descending",
        "unitless",
    ],
    "LPRM:LPRM_AMSRE_A_SOILM3.002:sm_c_error": [
        "Soil moisture uncertainty of LPRM AMSRE C-band",
        "percent",
    ],
    "LPRM:LPRM_AMSRE_A_SOILM3.002:sm_x_error": [
        "Soil moisture uncertainty of LPRM AMSRE X-band",
        "percent",
    ],
    "LPRM:LPRM_AMSRE_A_SOILM3.002:soil_moisture_c": [
        "Soil moisture, volumetric, from LPRM AMSRE C-band descending",
        "percent",
    ],
    "LPRM:LPRM_AMSRE_A_SOILM3.002:soil_moisture_x": [
        "Soil moisture, volumetric, from LPRM AMSRE X-band descending",
        "percent",
    ],
    "LPRM:LPRM_AMSRE_A_SOILM3.002:Ts": [
        "Skin temperature (2mm) from LPRM AMSRE descending",
        "K",
    ],
    "LPRM:LPRM_AMSRE_D_RZSM3.001:soilMoisture": [
        "Root Zone Soil Moisture from Palmer Water Balance Model",
        "percent",
    ],
    "LPRM:LPRM_AMSR2_A_SOILM3.001:soil_moisture_c1": [
        "Volumetric Soil Moisture from 6.9 GHZ, ascending",
        "percent",
    ],
    "LPRM:LPRM_AMSR2_D_SOILM3.001:soil_moisture_c1": [
        "Volumetric Soil Moisture from 6.9 GHZ, descending",
        "percent",
    ],
    "LPRM:LPRM_AMSR2_DS_A_SOILM3.001:soil_moisture_c1": [
        "Volumetric Soil Moisture from 6.9 GHZ, ascending",
        "percent",
    ],
    "LPRM:LPRM_AMSR2_DS_D_SOILM3.001:soil_moisture_c1": [
        "Volumetric Soil Moisture from 6.9 GHZ, descending",
        "percent",
    ],
    "LPRM:LPRM_TMI_DY_SOILM3.001:opt_depth_x": [
        "Optical depth from LPRM/TMI/TRMM X-band",
        "unitless",
    ],
    "LPRM:LPRM_TMI_DY_SOILM3.001:sm_x_error": [
        "Uncertainty of Soil Moisture in LPRM/TMI/TRMM X-band",
        "m3/m3",
    ],
    "LPRM:LPRM_TMI_DY_SOILM3.001:soil_moisture_x": [
        "Soil Moisture, Volumetric, from LPRM/TMI/TRMM X-band",
        "percent",
    ],
    "LPRM:LPRM_TMI_DY_SOILM3.001:ts": [
        "Skin temperature (2mm) from LPRM/TMI/TRMM",
        "K",
    ],
    "LPRM:LPRM_TMI_NT_SOILM3.001:opt_depth_x": [
        "Optical depth from LPRM/TMI/TRMM X-band",
        "unitless",
    ],
    "LPRM:LPRM_TMI_NT_SOILM3.001:sm_x_error": [
        "Uncertainty of Soil moisture in LPRM/TMI/TRMM X-band",
        "m3/m3",
    ],
    "LPRM:LPRM_TMI_NT_SOILM3.001:soil_moisture_x": [
        "Volumetric Soil Moisture from LPRM/TMI/TRMM X-band",
        "percent",
    ],
    "LPRM:LPRM_TMI_NT_SOILM3.001:ts": [
        "Skin temperature (2mm) from LPRM/TMI/TRMM",
        "K",
    ],
    "TRMM:TRMM_3B42.7:precipitation": ["Precipitation", "mm/hr"],
    "SMERGE:SMERGE_RZSM0_40CM_2.0:CCI_ano": [
        "CCI derived soil moisture anomalies of 0-40 cm layer",
        "m3/m3",
    ],
    "SMERGE:SMERGE_RZSM0_40CM_2.0:RZSM": [
        "Average soil moisture of 0-40 cm layer",
        "m3/m3",
    ],
    "GRACE:GRACEDADM_CLSM0125US_7D.2:gws_inst": [
        "Groundwater storage percentile",
        "percent",
    ],
    "GRACE:GRACEDADM_CLSM0125US_7D.2:rtzsm_inst": [
        "Root zone soil moisture percentile",
        "percent",
    ],
    "GRACE:GRACEDADM_CLSM0125US_7D.2:sfsm_inst": [
        "Surface soil moisture percentile",
        "percent",
    ],
    "MERRA:M2I1NXLFO.5124:QLML": ["Surface specific humidity, Instantaneous", "1"],
    "MERRA:M2I1NXLFO.5124:TLML": [
        "Surface air temperature over land, Instantaneous",
        "K",
    ],
    "MERRA:M2I1NXLFO.5124:SPEEDLML": ["Surface wind speed, Instantaneous", "m/s"],
    "MERRA:M2T1NXFLX.5124:ULML": ["Surface eastward wind, time average", "m/s"],
    "MERRA:M2T1NXFLX.5124:VLML": ["Surface northward wind, time average", "m/s"],
    "MERRA:M2T1NXLFO.5124:LWGAB": [
        "Surface absorbed longwave radiation, time average",
        "W/m2",
    ],
    "MERRA:M2T1NXLFO.5124:SWGDN": [
        "Incident shortwave radiation land, time average",
        "W/m2",
    ],
    "MERRA:MST1NXMLD.520:BASEFLOW": ["Baseflow", "kg/m2/s"],
    "MERRA:MST1NXMLD.520:LHLAND": ["Latent heat flux from land", "W/m2"],
    "MERRA:MST1NXMLD.520:PRECSNO": ["Surface snowfall", "kg/m2/s"],
    "MERRA:MST1NXMLD.520:PRECTOT": ["Total surface precipitation", "kg/m2/s"],
    "MERRA:MST1NXMLD.520:RUNOFF": ["Overland runoff", "kg/m2/s"],
    "MERRA:MST1NXMLD.520:SFMC": ["Top soil layer soil moisture content", "m3/m3"],
    "MERRA:MST1NXMLD.520:SHLAND": ["Sensible heat flux from land", "W/m2"],
    "MERRA:MST1NXMLD.520:TSOIL1": ["Soil temperature in layer 1", "K"],
}

# LDAS


class LDAS(object):
    def __init__(self, url, **query_params):
        query_params["type"] = "asc2"
        self.url = url
        self.query_params = query_params
        self.query_params["startDate"] = tsutils.parsedate(
            self.query_params["startDate"], strftime="%Y-%m-%dT%H"
        )
        self.query_params["endDate"] = tsutils.parsedate(
            self.query_params["endDate"], strftime="%Y-%m-%dT%H"
        )


# Function to make `resource` know about the new Daymet type.
# http://hydro1.gesdisc.eosdis.nasa.gov/daac-bin/access/timeseries.cgi?variable=GLDAS:GLDAS_NOAH025_3H.001:SOILM10-40cm&location=GEOM:POINT%28-99.875,%2031.125%29&startDate=2010-06-01T09&endDate=2015-05-04T21&type=asc2
#
# https://hydro1.gesdisc.eosdis.nasa.gov/daac-bin/access/timeseries.cgi?variable=NLDAS:NLDAS_FORA0125_H.002:APCPsfc&location=NLDAS:X304-Y071&startDate=2015-01-01T00&endDate=2015-06-20T23&type=asc2
#  http://hydro1.gesdisc.eosdis.nasa.gov/daac-bin/access/timeseries.cgi?variable=NLDAS:NLDAS_FORA0125_H.002:APCPsfc&location=NLDAS:X304-Y071&startDate=2015-01-01T00&endDate=2015-06-20T23&type=asc2


@resource.register(
    r"https://hydro1\.gesdisc\.eosdis\.nasa\.gov/daac-bin/access/timeseries\.cgi.*",
    priority=17,
)
def resource_ldas(uri, **kwargs):
    return LDAS(uri, **kwargs)


def _parse_ldas_dates(date, hour):
    try:
        return pd.to_datetime(date) + pd.to_timedelta(pd.np.int(hour[:-1]), "h")
    except (TypeError, ValueError):
        return pd.NaT


# Function to convert from LDAS type to pd.DataFrame


@convert.register(pd.DataFrame, LDAS)
def ldas_to_df(data, **kwargs):
    req = requests.get(data.url, params=data.query_params)
    if os.path.exists("debug_tsgettoolbox"):
        logging.warning(req.url)
    req.raise_for_status()

    df = pd.read_table(
        BytesIO(req.content),
        skiprows=40,
        header=None,
        index_col=None,
        delim_whitespace=True,
        na_values=[-9999, -9999.0],
    )
    df.drop(df.index[-1], axis="rows", inplace=True)
    if len(df.columns) == 3:
        df["dt"] = df[0] + "T" + df[1]
        df["dt"] = pd.to_datetime(df["dt"])
        df.set_index("dt", inplace=True)
        df.drop([0, 1], axis="columns", inplace=True)
    else:
        df[0] = pd.to_datetime(df[0])
        df.set_index(0, inplace=True)
    variable_name = data.query_params["variable"].split(":")[-1]
    unit = _UNITS_MAP[data.query_params["variable"]][1]
    df.columns = ["{0}:{1}".format(variable_name, unit)]
    df.index.name = "Datetime:UTC"
    try:
        return df.tz_localize("UTC")
    except TypeError:  # Already UTC
        return df


if __name__ == "__main__":
    # ?variable=GLDAS:GLDAS_NOAH025_3H.001:SOILM10-40cm&
    # location=GEOM:POINT%28-99.875,%2031.125%29&
    # startDate=2010-06-01T09&endDate=2015-05-04T21&type=asc2
    #
    for key in _UNITS_MAP:
        try:
            r = resource(
                r"https://hydro1.gesdisc.eosdis.nasa.gov/daac-bin/access/timeseries.cgi",
                variable=key,
                location="GEOM:POINT(-100, 31)",
                startDate="2013-06-01T09",
                endDate="2014-05-04T21",
            )

            as_df = odo(r, pd.DataFrame)
        except:
            r = resource(
                r"https://hydro1.gesdisc.eosdis.nasa.gov/daac-bin/access/timeseries.cgi",
                variable=key,
                location="GEOM:POINT(-100, 31)",
                startDate="2002-06-01T09",
                endDate="2003-05-04T21",
            )

            as_df = odo(r, pd.DataFrame)
        print("LDAS", key)
        print(as_df)
#
#     r = resource(
#         r'https://hydro1.gesdisc.eosdis.nasa.gov/daac-bin/access/timeseries.cgi',
#         variable='GLDAS:GLDAS_NOAH025_3H.001:SOILM10-40cm',
#         location='GEOM:POINT(104.2, 35.86)',
#         startDate='2016-01-01T00',
#         endDate='2016-12-01T00'
#     )
#
#     as_df = odo(r, pd.DataFrame)
#     print('LDAS TEST')
#     print(as_df)
#
#     r = resource(
#         r'https://hydro1.gesdisc.eosdis.nasa.gov/daac-bin/access/timeseries.cgi',
#         variable='GLDAS2:GLDAS_NOAH025_3H_v2.1:SOILM10-40cm',
#         location='GEOM:POINT(104.2, 35.86)',
#         startDate='5 years ago',
#         endDate='4 years ago'
#     )
#
#     as_df = odo(r, pd.DataFrame)
#     print('LDAS TEST')
#     print(as_df)



"""
    ulmo.nasa.daymet.core
    ~~~~~~~~~~~~~~

    This module provides direct access to `NASA EARTHDATA ORNL DAAC 
    Daymet`_ web services.


    .. .. _NASA EARTHDATA ORNL DAAC Daymet: https://daymet.ornl.gov/dataaccess.html

"""
from future import standard_library

standard_library.install_aliases()
from builtins import str
from past.builtins import basestring
import contextlib
import io
import datetime
import time
import logging

import requests
import pandas as pd

from tsgettoolbox.ulmo import util

VARIABLES = {
    "tmax": "maximum temperature",
    "tmin": "minimum temperature",
    "srad": "shortwave radiation",
    "vp": "vapor pressure",
    "swe": "snow-water equivalent",
    "prcp": "precipitation",
    "dayl": "daylength",
}
MIN_YEAR = 1980
MAX_Year = int(time.strftime("%Y")) - 1
MIN_LAT = 14.5
MAX_LAT = 52.0
MIN_LON = -131.0
MAX_LON = -53.0


DAYMET_SINGLEPIXEL_URL = "https://daymet.ornl.gov/data/send/saveData?lat={lat}&lon={lon}&measuredParams={vars}"

# configure logging
LOG_FORMAT = "%(message)s"
logging.basicConfig(format=LOG_FORMAT)
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def get_variables():
    """retrieve a list of variables available

    Parameters
    ----------
    None

    Returns
    -------
        dictionary of variables with variable abreviations as keys 
        and description as values
    """
    return VARIABLES


def get_daymet_singlepixel(
    latitude,
    longitude,
    variables=["tmax", "tmin", "prcp"],
    years=None,
    as_dataframe=True,
):
    """Fetches a time series of climate variables from the DAYMET single pixel extraction


    Parameters
    ----------
    latitude: float
        The latitude (WGS84), value between 52.0 and 14.5.
    longitude: float
        The longitude (WGS84), value between -131.0 and -53.0.
    variables : List of str
        Daymet parameters to fetch. 
        Available options:
            ``tmax`` - maximum temperature
            ``tmin`` - minimum temperature
            ``srad`` - shortwave radiation
            ``vp`` - vapor pressure
            ``swe`` - snow-water equivalent
            ``prcp`` - precipitation
            ``dayl`` - daylength
        default = ['tmax', 'tmin', 'prcp']
    years: list of int
        List of years to return. 
        Daymet version 2 available 1980 to the latest full calendar year.
        If ``None`` (default), all years will be returned
    as_dataframe : ``True`` (default) or ``False``
        if ``True`` return pandas dataframe
        if ``False`` return open file with contents in csv format

    
    Returns
    -------
    single_pixel_timeseries : pandas dataframe or csv filename
    """
    _check_coordinates(latitude, longitude)
    _check_variables(variables)
    if not years is None:
        _check_years(years)

    url_params = {"lat": latitude, "lon": longitude, "vars": _as_str(variables)}
    if years:
        url_params["years"] = _as_str(years)

    url = _get_service_url(url_params)
    log.info(
        "making request for latitude, longitude: {}, {}".format(latitude, longitude)
    )
    df = pd.read_csv(url, header=6)
    df.year, df.yday = df.year.astype("int"), df.yday.astype("int")
    df.index = pd.to_datetime(
        df.year.astype("str") + "-" + df.yday.astype("str"), format="%Y-%j"
    )
    df.columns = [c[: c.index("(")].strip() if "(" in c else c for c in df.columns]
    if as_dataframe:
        return df
    else:
        results = {}
        for key in df.columns:
            if key not in ["yday", "year"]:
                results[key] = dict(zip(df[key].index.format(), df[key]))
        return results


def _check_variables(variables):
    """make sure all variables are in list
    """
    bad_variables = [v for v in variables if v not in VARIABLES.keys()]
    if bad_variables:

        raise ValueError(
            "the variable(s) provided ('{}') not\none of available options: '{}'".format(
                "', '".join(bad_variables), str(VARIABLES.keys())[2:-2]
            )
        )


def _check_years(years):
    """make sure all years are in available year range
    """
    bad_years = [str(year) for year in years if not MIN_YEAR <= year <= MAX_Year]
    if bad_years:
        raise ValueError(
            "the year(s) provided ({}) \nnot in available timerange ({}-{})".format(
                ", ".join(bad_years), MIN_YEAR, MAX_Year
            )
        )


def _check_coordinates(lat, lon):
    """make sure the passed coordinates are in the available data range
    """
    bad_lat = not MIN_LAT <= lat <= MAX_LAT
    bad_lon = not MIN_LON <= lon <= MAX_LON

    if bad_lat or bad_lon:
        err_msg = "The specified latitude, longitude pair ({}, {})".format(lat, lon)
        err_msg += "\nis not in the available range of data:\n"
        err_msg += "\tLatitude = [{} - {}]".format(MIN_LAT, MAX_LAT)
        err_msg += "\n\tLongitude = [{} - {}]".format(MIN_LON, MAX_LON)
        raise ValueError(err_msg)


def _as_str(arg):
    """if arg is a list, convert to comma delimited string
    """
    if isinstance(arg, basestring):
        return arg
    else:
        return ",".join([str(a) for a in arg])


def _get_service_url(url_params):
    """return formatted url
    """
    url = DAYMET_SINGLEPIXEL_URL.format(**url_params)

    if "years" in url_params:
        url += "&year={}".format(url_params["years"])
    return url



"""
Access hydrologic and climate data from the `Lower Colorado River Authority`_.

Data for the Colorado River Basin (Texas) provided by the `Lower Colorado River
Authority`_ `Hydromet`_ web site and web service.

.. _Lower Colorado River Authority: http://www.lcra.org
.. _Hydromet: http://hydromet.lcra.org
"""
from __future__ import print_function

from builtins import str

import pandas as pd

from tsgettoolbox.ulmo.lcra.hydromet.core import get_current_data
from tsgettoolbox.ulmo.lcra.hydromet.core import get_site_data


def ulmo_df(
    site_code, parameter_code, start_date=None, end_date=None, dam_site_location="head"
):
    if parameter_code.lower() in ["upperbasin", "lowerbasin"]:
        # def get_current_data(service, as_geojson=False):
        df = pd.DataFrame(get_current_data("get" + parameter_code))
        return df.set_index("location")

    # def get_site_data(site_code, parameter_code, as_dataframe=True,
    #     start_date=None, end_date=None, dam_site_location='head'):
    return get_site_data(
        str(site_code),
        parameter_code,
        start_date=pd.to_datetime(start_date),
        end_date=pd.to_datetime(end_date),
        dam_site_location=dam_site_location,
    )


if __name__ == "__main__":
    #    import time
    #
    #    r = ulmo_df('blah',
    #                'upperbasin')
    #
    #    print('UB EVERYTHING')
    #    print(r)
    #
    #    time.sleep(20)

    r = ulmo_df(4598, "stage", start_date="2015-11-04", end_date="2015-12-05")

    print("UB EVERYTHING")
    print(r)



from __future__ import absolute_import, division, print_function

import os
from itertools import product
import pandas as pd
import pytest

sa = pytest.importorskip("sqlalchemy")

from datashape import dshape, discover
from tsgettoolbox.odo import resource, odo
from tsgettoolbox.odo.utils import tmpfile, filetext


ds = dshape("var *  {a: int32, b: int32}")
data = [(1, 2), (10, 20), (100, 200)]


@pytest.yield_fixture
def csv():
    with tmpfile("csv") as filename:
        yield odo(data, filename, dshape=ds, has_header=False)


def test_simple_into(csv):
    tbl = "testtable"
    with tmpfile("db") as filename:
        engine = resource("sqlite:///" + filename)
        t = resource("sqlite:///" + filename + "::" + tbl, dshape=ds)

        odo(csv, t, dshape=ds)
        conn = engine.raw_connection()
        cursor = conn.cursor()
        cursor.execute(
            """SELECT
                        name
                       FROM
                        sqlite_master
                       WHERE type='table' and name='{0}';""".format(
                tbl
            )
        )

        sqlite_tbl_names = cursor.fetchall()
        conn.close()

        assert sqlite_tbl_names[0][0] == tbl
        assert odo(t, list) == data


def test_csv_with_header():
    with tmpfile("db") as dbfilename:
        with filetext("a,b\n1,2\n3,4", extension="csv") as csvfilename:
            t = odo(csvfilename, "sqlite:///%s::mytable" % dbfilename)
            assert discover(t) == dshape("var * {a: int64, b: int64}")
            assert odo(t, set) == set([(1, 2), (3, 4)])


def test_csv_infer_header():
    with tmpfile("db") as dbfilename:
        with filetext("a,b\n1,2\n3,4", extension="csv") as csvfilename:
            t = odo(csvfilename, "sqlite:///%s::mytable" % dbfilename)
            assert discover(t) == dshape("var * {a: int64, b: int64}")
            assert odo(t, set) == set([(1, 2), (3, 4)])


@pytest.mark.parametrize(["sep", "header"], product([",", "|", "\t"], [True, False]))
def test_sqlite_to_csv(sep, header):
    with tmpfile("db") as dbfilename:
        with filetext("a,b\n1,2\n3,4", extension="csv") as csvfilename:
            t = odo(csvfilename, "sqlite:///%s::mytable" % dbfilename)

        with tmpfile(".csv") as fn:
            odo(t, fn, header=header, delimiter=sep)
            with open(fn, "rt") as f:
                lines = f.readlines()
            expected = [
                tuple(map(int, row))
                for row in map(lambda x: x.split(sep), lines[header:])
            ]
            assert (
                odo(fn, list, delimiter=sep, has_header=header, dshape=discover(t))
                == expected
            )


def test_different_encoding():
    encoding = "latin1"
    with tmpfile("db") as db:
        sql = odo(
            os.path.join(os.path.dirname(__file__), "encoding.csv"),
            "sqlite:///%s::t" % db,
            encoding=encoding,
        )
        result = odo(sql, list)

    NULL = u""

    expected = [
        (u"1958.001.500131-1A", 1, NULL, NULL, 899),
        (u"1958.001.500156-6", 1, NULL, NULL, 899),
        (u"1958.001.500162-1", 1, NULL, NULL, 899),
        (u"1958.001.500204-2", 1, NULL, NULL, 899),
        (u"1958.001.500204-2A", 1, NULL, NULL, 899),
        (u"1958.001.500204-2B", 1, NULL, NULL, 899),
        (u"1958.001.500223-6", 1, NULL, NULL, 9610),
        (u"1958.001.500233-9", 1, NULL, NULL, 4703),
        (u"1909.017.000018-3", 1, 30.0, u"sumaria", 899),
    ]
    assert result == expected


@pytest.yield_fixture
def quoted_sql():
    with tmpfile(".db") as db:
        try:
            yield resource("sqlite:///%s::foo bar" % db, dshape=ds)
        except sa.exc.OperationalError as e:
            pytest.skip(str(e))


@pytest.mark.xfail(
    raises=sa.exc.DatabaseError,
    reason="How do you use a quoted table name with the SQLite .import command?",
)
def test_quoted_name(csv, quoted_sql):
    with tmpfile("csv") as filename:
        csv = odo(data, filename, dshape=ds, has_header=True)
        s = odo(csv, quoted_sql)
        t = odo(csv, list)
        assert sorted(odo(s, list)) == sorted(t)


def test_different_encoding_to_csv():
    with tmpfile("db") as dbfilename:
        with filetext("a,b\n1,2\n3,4", extension="csv") as csvfilename:
            t = odo(
                csvfilename, "sqlite:///%s::mytable" % dbfilename, encoding="latin1"
            )
            with tmpfile(".csv") as fn:
                with pytest.raises(ValueError):
                    odo(t, fn, encoding="latin1")


def test_send_parameterized_query_to_csv():
    with tmpfile("db") as dbfilename:
        with filetext("a,b\n1,2\n3,4", extension="csv") as csvfilename:
            t = odo(csvfilename, "sqlite:///%s::mytable" % dbfilename)
        with tmpfile(".csv") as fn:
            q = t.select(t.c.a == 1)
            r = odo(q, fn)
            assert sorted(odo(q, list)) == sorted(odo(r, list))


def test_df_to_in_memory_db():
    df = pd.DataFrame([[1, 2], [3, 4]], columns=list("ab"))
    tbl = odo(df, "sqlite:///:memory:::tbl")
    pd.util.testing.assert_frame_equal(odo(tbl, pd.DataFrame), df)



from __future__ import absolute_import, division, print_function

import pytest

sa = pytest.importorskip("sqlalchemy")
psycopg2 = pytest.importorskip("psycopg2")

import datetime
import os
import itertools
import shutil

from datashape import dshape
from datashape.util.testing import assert_dshape_equal
import numpy as np
import pandas as pd

from tsgettoolbox.odo.backends.csv import CSV
from tsgettoolbox.odo.backends.sql import select_to_base
from tsgettoolbox.odo import odo, into, resource, drop, discover, convert
from tsgettoolbox.odo.utils import assert_allclose, tmpfile


@pytest.fixture(scope="module")
def pg_ip():
    return os.environ.get("POSTGRES_IP", "localhost")


skip_no_rw_loc = pytest.mark.skipif(
    pg_ip() != "localhost",
    reason=("Skipping tests with remote PG server (%s)" % pg_ip()),
)

names = ("tbl%d" % i for i in itertools.count())
new_schema = object()
data = [(1, 2), (10, 20), (100, 200)]
null_data = [(1, None), (10, 20), (100, 200)]


@pytest.fixture(scope="module")
def tmpdir():
    return os.environ.get("POSTGRES_TMP_DIR", None)


@pytest.yield_fixture
def csv(tmpdir):
    s = "\n".join(",".join(map(str, row)) for row in data).encode("utf8")
    with tmpfile(".csv", dir=tmpdir) as fn:
        with open(fn, "wb") as f:
            f.write(s)
        yield CSV(fn)


@pytest.yield_fixture
def encoding_csv(tmpdir):
    path = os.path.join(os.path.dirname(__file__), "encoding.csv")
    with tmpfile(".csv", dir=tmpdir) as fn:
        with open(fn, "wb") as f, open(path, "r") as g:
            f.write(g.read().encode("latin1"))
        yield CSV(fn)


@pytest.yield_fixture
def complex_csv(tmpdir):
    path = os.path.join(os.path.dirname(__file__), "dummydata.csv")
    with tmpfile(".csv", dir=tmpdir) as fn:
        shutil.copy(path, fn)
        os.chmod(fn, 0o777)
        yield CSV(fn, has_header=True)


@pytest.fixture
def url(pg_ip):
    return "postgresql://postgres@{}/test::{}".format(pg_ip, next(names))


def sql_fixture(dshape, schema=None):
    @pytest.yield_fixture(params=(True, False))
    def fixture(request, url):
        try:
            t = resource(
                url,
                dshape=dshape,
                schema=next(names) if schema is new_schema else schema,
            )
        except sa.exc.OperationalError as e:
            pytest.skip(str(e))
        else:
            try:
                bind = t.bind
                if not request.param:
                    t.metadata.bind = None
                yield t, bind
            finally:
                t.metadata.bind = bind
                drop(t)
                if t.schema:
                    bind.execute(sa.sql.ddl.DropSchema(t.schema))

    return fixture


sql = sql_fixture("var * {a: int32, b: ?int32}")
sql_with_schema = sql_fixture("var * {a: int32, b: ?int32}", new_schema)
sql_with_ugly_schema = sql_fixture("var * {a: int32, b: ?int32}", "foo.b.ar")
complex_sql = sql_fixture(
    """
    var * {
        Name: string, RegistrationDate: date, ZipCode: int32, Consts: float64
    }
"""
)
sql_with_floats = sql_fixture("var * {a: float64, b: ?float64}")
sql_with_dts = sql_fixture("var * {a: datetime, b: ?datetime}")
sql_with_datelikes = sql_fixture(
    """
    var * {
        date: date,
        option_date: ?date,
        datetime: datetime,
        option_datetime: ?datetime,
    }
"""
)
sql_with_strings = sql_fixture(
    """
    var * {non_optional: string, optional: ?string}
"""
)


@pytest.yield_fixture
def quoted_sql(pg_ip, csv):
    url = "postgresql://postgres@{}/test::foo bar".format(pg_ip)
    try:
        t = resource(url, dshape=discover(csv))
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)


@skip_no_rw_loc
def test_simple_into(csv, sql):
    sql, bind = sql
    into(sql, csv, dshape=discover(sql), bind=bind)
    assert into(list, sql, bind=bind) == data


@skip_no_rw_loc
def test_append(csv, sql):
    sql, bind = sql
    into(sql, csv, bind=bind)
    assert into(list, sql, bind=bind) == data

    into(sql, csv, bind=bind)
    assert into(list, sql, bind=bind) == data + data


def test_tryexcept_into(csv, sql):
    sql, bind = sql
    with pytest.raises(psycopg2.NotSupportedError):
        # uses multi-byte character
        into(sql, csv, quotechar="alpha", bind=bind)


@skip_no_rw_loc
def test_no_header_no_columns(csv, sql):
    sql, bind = sql
    into(sql, csv, bind=bind, dshape=discover(sql))
    assert into(list, sql, bind=bind) == data


@skip_no_rw_loc
def test_complex_into(complex_csv, complex_sql):
    complex_sql, bind = complex_sql
    # data from: http://dummydata.me/generate
    into(complex_sql, complex_csv, dshape=discover(complex_sql), bind=bind)
    assert_allclose(into(list, complex_sql, bind=bind), into(list, complex_csv))


@skip_no_rw_loc
def test_sql_to_csv(sql, csv, tmpdir):
    sql, bind = sql
    sql = odo(csv, sql, bind=bind)
    with tmpfile(".csv", dir=tmpdir) as fn:
        csv = odo(sql, fn, bind=bind)
        assert odo(csv, list) == data
        assert discover(csv).measure.names == discover(sql).measure.names


@skip_no_rw_loc
def test_sql_select_to_csv(sql, csv, tmpdir):
    sql, bind = sql
    sql = odo(csv, sql, bind=bind)
    query = sa.select([sql.c.a])
    with tmpfile(".csv", dir=tmpdir) as fn:
        csv = odo(query, fn, bind=bind)
        assert odo(csv, list) == [(x,) for x, _ in data]


def test_invalid_escapechar(sql, csv):
    sql, bind = sql
    with pytest.raises(ValueError):
        odo(csv, sql, escapechar="12", bind=bind)

    with pytest.raises(ValueError):
        odo(csv, sql, escapechar="", bind=bind)


@skip_no_rw_loc
def test_csv_output_is_not_quoted_by_default(sql, csv, tmpdir):
    sql, bind = sql
    sql = odo(csv, sql, bind=bind)
    expected = "a,b\n1,2\n10,20\n100,200\n"
    with tmpfile(".csv", dir=tmpdir) as fn:
        csv = odo(sql, fn, bind=bind)
        with open(fn, "rt") as f:
            result = f.read()
        assert result == expected


@skip_no_rw_loc
def test_na_value(sql, csv, tmpdir):
    sql, bind = sql
    sql = odo(null_data, sql, bind=bind)
    with tmpfile(".csv", dir=tmpdir) as fn:
        csv = odo(sql, fn, na_value="NA", bind=bind)
        with open(csv.path, "rt") as f:
            raw = f.read()
    assert raw == "a,b\n1,NA\n10,20\n100,200\n"


def test_different_encoding(url, encoding_csv):
    try:
        sql = odo(encoding_csv, url, encoding="latin1")
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            result = odo(sql, list)
            expected = [
                (u"1958.001.500131-1A", 1, None, u"", 899),
                (u"1958.001.500156-6", 1, None, u"", 899),
                (u"1958.001.500162-1", 1, None, u"", 899),
                (u"1958.001.500204-2", 1, None, u"", 899),
                (u"1958.001.500204-2A", 1, None, u"", 899),
                (u"1958.001.500204-2B", 1, None, u"", 899),
                (u"1958.001.500223-6", 1, None, u"", 9610),
                (u"1958.001.500233-9", 1, None, u"", 4703),
                (u"1909.017.000018-3", 1, 30.0, u"sumaria", 899),
            ]
            assert result == expected
        finally:
            drop(sql)


@skip_no_rw_loc
def test_schema(csv, sql_with_schema):
    sql_with_schema, bind = sql_with_schema
    assert odo(odo(csv, sql_with_schema, bind=bind), list, bind=bind) == data


@skip_no_rw_loc
def test_ugly_schema(csv, sql_with_ugly_schema):
    sql_with_ugly_schema, bind = sql_with_ugly_schema
    assert odo(odo(csv, sql_with_ugly_schema, bind=bind), list, bind=bind) == data


def test_schema_discover(sql_with_schema):
    sql_with_schema, bind = sql_with_schema
    sql_with_schema.metadata.bind = bind
    meta = discover(sql_with_schema.metadata)
    assert meta == dshape("{%s: var * {a: int32, b: ?int32}}" % sql_with_schema.name)


@skip_no_rw_loc
def test_quoted_name(quoted_sql, csv):
    s = odo(csv, quoted_sql)
    t = odo(csv, list)
    assert sorted(odo(s, list)) == sorted(t)


def test_path_of_reduction(sql):
    sql, bind = sql
    result = list(convert.path(sa.select([sa.func.sum(sql.c.a)]), float))
    expected = [(sa.sql.Select, float, select_to_base, 200)]
    assert result == expected


def test_drop_reflects_database_state(url):
    data = list(zip(range(5), range(1, 6)))

    t = odo(data, url, dshape="var * {A: int64, B: int64}")
    assert t.exists()
    assert resource(url).exists()

    drop(url)
    with pytest.raises(ValueError):
        resource(url)  # Table doesn't exist and no dshape


def test_nan_stays_nan(sql_with_floats):
    sql_with_floats, bind = sql_with_floats

    df = pd.DataFrame({"a": [np.nan, 1, 2], "b": [3, np.nan, 4]})
    odo(df, sql_with_floats, bind=bind)
    rehydrated = odo(sql_with_floats, pd.DataFrame, bind=bind)
    pd.util.testing.assert_frame_equal(rehydrated.sort_index(axis=1), df)

    nulls_query = sa.select(sql_with_floats.c).where(
        sql_with_floats.c.a.is_(None) | sql_with_floats.c.b.is_(None)
    )
    if bind is None:
        nulls = nulls_query.execute()
    else:
        nulls = bind.execute(nulls_query)
    assert not nulls.fetchall()


def test_nat_to_null(sql_with_dts):
    sql_with_dts, bind = sql_with_dts

    df = pd.DataFrame(
        {
            "a": pd.to_datetime(["2014-01-01", "2014-01-02"]),
            "b": pd.to_datetime(["2014-01-01", "nat"]),
        }
    )
    odo(df, sql_with_dts, bind=bind)
    rehydrated = odo(sql_with_dts, pd.DataFrame, bind=bind)
    pd.util.testing.assert_frame_equal(rehydrated.sort_index(axis=1), df)

    nulls_query = sa.select(sql_with_dts.c).where(
        sql_with_dts.c.a.is_(None) | sql_with_dts.c.b.is_(None)
    )
    if bind is None:
        nulls = nulls_query.execute()
    else:
        nulls = bind.execute(nulls_query)
    assert nulls.fetchall() == [(pd.Timestamp("2014-01-02"), None)]


def test_to_dataframe(sql):
    sql, bind = sql
    insert_query = sql.insert().values([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    if bind is None:
        insert_query.execute()
    else:
        bind.execute(insert_query)

    df = odo(sql, pd.DataFrame, bind=bind)
    expected = pd.DataFrame(
        np.array([(1, 2), (3, 4)], dtype=[("a", "int32"), ("b", "float64")])
    )
    pd.util.testing.assert_frame_equal(df, expected)


def test_to_dataframe_datelike(sql_with_datelikes):
    sql_with_datelikes, bind = sql_with_datelikes
    date_0 = datetime.date(2014, 1, 1)
    date_1 = datetime.date(2014, 1, 2)
    datetime_0 = datetime.datetime(2014, 1, 1)
    datetime_1 = datetime.datetime(2014, 1, 2)
    insert_query = sql_with_datelikes.insert().values(
        [
            {
                "date": date_0,
                "option_date": date_0,
                "datetime": datetime_0,
                "option_datetime": datetime_0,
            },
            {
                "date": date_1,
                "option_date": date_1,
                "datetime": datetime_1,
                "option_datetime": datetime_1,
            },
        ]
    )
    if bind is None:
        insert_query.execute()
    else:
        bind.execute(insert_query)

    df = odo(sql_with_datelikes, pd.DataFrame, bind=bind)
    expected = pd.DataFrame(
        [[datetime_0] * 4, [datetime_1] * 4],
        columns=["date", "option_date", "datetime", "option_datetime"],
    )
    pd.util.testing.assert_frame_equal(df, expected)


def test_to_dataframe_strings(sql_with_strings):
    sql_with_strings, bind = sql_with_strings

    insert_query = sql_with_strings.insert().values(
        [
            {"non_optional": "ayy", "optional": 'hello "world"'},
            {"non_optional": "lmao", "optional": None},
        ]
    )
    if bind is None:
        insert_query.execute()
    else:
        bind.execute(insert_query)

    df = odo(sql_with_strings, pd.DataFrame, bind=bind)
    expected = pd.DataFrame(
        [["ayy", 'hello "world"'], ["lmao", None]], columns=["non_optional", "optional"]
    )
    pd.util.testing.assert_frame_equal(df, expected)


def test_to_dataframe_modulo(sql):
    sql, bind = sql

    insert_query = sql.insert().values([{"a": n} for n in range(10)])
    if bind is None:
        insert_query.execute()
    else:
        bind.execute(insert_query)

    lookup_query = sa.select([(sql.c.a % 5).label("a")])
    result = odo(lookup_query, pd.DataFrame, bind=bind, dshape="var * {a: int32}")
    expected = pd.DataFrame({"a": np.arange(10) % 5}, dtype="int32")
    pd.util.testing.assert_frame_equal(result, expected)


def test_from_dataframe_strings(sql_with_strings):
    sql_with_strings, bind = sql_with_strings

    input_ = pd.DataFrame(
        [["ayy", 'hello "world"'], ["lmao", None]], columns=["non_optional", "optional"]
    )
    odo(input_, sql_with_strings, bind=bind)
    output = odo(sql_with_strings, pd.DataFrame, bind=bind)
    pd.util.testing.assert_frame_equal(output, input_)


def test_float_dtype(sql_with_floats):
    sql_with_floats, bind = sql_with_floats

    expected = dshape("var * {a: float64, b: ?float64}")

    assert_dshape_equal(discover(sql_with_floats), expected)

    # Also check that reflection from the database returns expected dshape.
    assert_dshape_equal(discover(bind).subshape[sql_with_floats.name], expected)



import isodate

from lxml import etree
from past.builtins import basestring

from tsgettoolbox.ulmo import util


def parse_site_values(content_io, namespace, query_isodate=None, methods=None):
    """parses values out of a waterml file; content_io should be a file-like object"""
    data_dict = {}
    metadata_elements = [
        # (element name, name of collection,
        #   key from element dict to use as for a key in the collections dict)
        ("censorCode", "censor_codes", "censor_code"),
        ("method", "methods", "id"),
        ("offset", "offsets", "id"),
        ("qualifier", "qualifiers", "id"),
        ("qualityControlLevel", "quality_control_levels", "id"),
        ("source", "sources", "id"),
    ]
    for (event, ele) in etree.iterparse(content_io):
        if ele.tag == namespace + "timeSeries":
            source_info_element = ele.find(namespace + "sourceInfo")
            site_info = _parse_site_info(source_info_element, namespace)
            var_element = ele.find(namespace + "variable")
            variable = _parse_variable(var_element, namespace)
            values_elements = ele.findall(namespace + "values")
            code = variable["code"]
            if isinstance(methods, basestring):
                method = methods
            elif isinstance(methods, dict):
                method = methods.get(code, None)
            else:
                method = None
            if "statistic" in variable:
                code += ":" + variable["statistic"]["code"]

            if method is None:
                if len(values_elements) > 1:
                    raise ValueError(
                        "found more than one method for %s. need to specify"
                        'specify code or "all".' % variable["code"]
                    )
                values_element = values_elements[0]
                values = _parse_values(values_element, namespace)
                data_dict[code] = {"site": site_info, "variable": variable}
                data_dict[code].update({"values": values})
                metadata = _parse_metadata(values_element, metadata_elements, namespace)
                data_dict[code].update(metadata)
                if query_isodate:
                    data_dict[code]["last_refresh"] = query_isodate
            elif method == "all":
                for values_element in values_elements:
                    values = _parse_values(values_element, namespace)
                    metadata = _parse_metadata(
                        values_element, metadata_elements, namespace
                    )
                    if len(values_elements) > 1:
                        updated_code = (
                            code
                            + ":"
                            + str(list(metadata["methods"].values())[0]["id"])
                        )
                    else:
                        updated_code = code
                    data_dict[updated_code] = {
                        "site": site_info.copy(),
                        "variable": variable,
                    }
                    data_dict[updated_code].update({"values": values})
                    data_dict[updated_code].update(metadata)
                    if query_isodate:
                        data_dict[updated_code]["last_refresh"] = query_isodate
            else:
                for values_element in values_elements:
                    if (
                        values_element.find(
                            namespace + 'method[@methodID="%s"]' % method
                        )
                        is not None
                    ):
                        values = _parse_values(values_element, namespace)
                        metadata = _parse_metadata(
                            values_element, metadata_elements, namespace
                        )
                        data_dict[code] = {"site": site_info, "variable": variable}
                        data_dict[code].update({"values": values})
                        data_dict[code].update(metadata)
                        if query_isodate:
                            data_dict[code]["last_refresh"] = query_isodate

    return data_dict


def parse_site_infos(content_io, namespace, site_info_names):
    """parses information contained in site info elements out of a waterml file;
    content_io should be a file-like object
    """
    site_infos = {}
    for site_info_name in site_info_names:
        content_io.seek(0)
        site_info_elements = [
            element
            for (event, element) in etree.iterparse(content_io)
            if element.tag == namespace + site_info_name
        ]
        site_info_dicts = [
            _parse_site_info(site_info_element, namespace)
            for site_info_element in site_info_elements
        ]
        site_infos.update(dict([(d["code"], d) for d in site_info_dicts]))
    return site_infos


def parse_sites(content_io, namespace):
    """parses information contained in site elements (including seriesCatalogs)
    out of a waterml file; content_io should be a file-like object
    """
    content_io.seek(0)
    site_elements = [
        ele
        for (event, ele) in etree.iterparse(content_io)
        if ele.tag == namespace + "site"
    ]
    site_dicts = [
        _parse_site(site_element, namespace) for site_element in site_elements
    ]
    sites = dict([(site_dict["code"], site_dict) for site_dict in site_dicts])
    return sites


def parse_variables(content_io, namespace):
    """parses information contained in variables elements out of a waterml file;
    content_io should be a file-like object
    """
    content_io.seek(0)
    variable_elements = [
        element
        for (event, element) in etree.iterparse(content_io)
        if element.tag == namespace + "variable"
    ]
    variable_dicts = [
        _parse_variable(variable_element, namespace)
        for variable_element in variable_elements
    ]
    variables = dict(
        [(variable_dict["code"], variable_dict) for variable_dict in variable_dicts]
    )
    return variables


def _element_dict(element, exclude_children=None, prepend_attributes=True):
    """converts an element to a dict representation with CamelCase tag names and
    attributes converted to underscores; this is a generic converter for cases
    where special parsing isn't necessary.  In most cases you will want to
    update with this dict. If prepend_element_name is True (default), then
    attributes and children will be prepended with the parent element's tag
    name.

    Note: does not handle sibling elements
    """
    if element is None:
        return {}

    if exclude_children is None:
        exclude_children = []

    element_dict = {}
    element_name = util.camel_to_underscore(element.tag.split("}")[-1])

    if len(element) == 0 and not element.text is None:
        element_dict[element_name] = element.text

    element_dict.update(
        dict(
            [
                (
                    _element_dict_attribute_name(
                        key, element_name, prepend_element_name=prepend_attributes
                    ),
                    value,
                )
                for key, value in element.attrib.items()
                if value.split(":")[0] not in ["xsd", "xsi"]
            ]
        )
    )

    for child in element.iterchildren():
        if not child.tag.split("}")[-1] in exclude_children:
            element_dict.update(_element_dict(child))

    return element_dict


def _element_dict_attribute_name(
    attribute_name, element_name, prepend_element_name=True
):
    attribute_only = util.camel_to_underscore(attribute_name.split("}")[-1])
    if attribute_only.startswith(element_name) or not prepend_element_name:
        return attribute_only
    else:
        return element_name + "_" + attribute_only


def _find_unit(element, namespace):
    unit_element = element.find(namespace + "unit")
    if unit_element is None:
        unit_element = element.find(namespace + "units")
    return unit_element


def _parse_datetime(datetime_str):
    """returns an iso 8601 datetime string; USGS returns fractions of a second
    which are usually all 0s. ISO 8601 does not limit the number of decimal
    places but we have to cut them off at some point
    """
    # XXX: this could be sped up if need be
    # XXX: also, we need to document that we are throwing away fractions of
    #     seconds
    return isodate.datetime_isoformat(isodate.parse_datetime(datetime_str))


def _parse_geog_location(geog_location, namespace):
    """returns a dict representation of a geogLocation etree element"""
    return_dict = {
        "latitude": geog_location.find(namespace + "latitude").text,
        "longitude": geog_location.find(namespace + "longitude").text,
    }

    srs = geog_location.attrib.get("srs")
    if not srs is None:
        return_dict["srs"] = srs

    return return_dict


def _parse_metadata(values_element, metadata_elements, namespace):
    metadata = {}
    for tag, collection_name, key in metadata_elements:
        underscored_tag = util.camel_to_underscore(tag)
        collection = [
            _scrub_prefix(_element_dict(element, namespace), underscored_tag)
            for element in values_element.findall(namespace + tag)
        ]
        if len([x for x in collection if len(x)]):
            collection_dict = dict(
                [(item[key], item) for item in collection if key in item]
            )
            metadata[collection_name] = collection_dict
    return metadata


def _parse_method(method, namespace):
    return _element_dict(method, namespace, prepend_attributes=False)


def _parse_series(series, namespace):
    include_elements = [
        "method",
        "Method",
        "source",
        "Source",
        "QualityControlLevel",
        "qualityControlLevel",
        "variableTimeInterval",
        "valueCount",
    ]
    series_dict = {}

    variable_element = series.find(namespace + "variable")
    series_dict["variable"] = _parse_variable(variable_element, namespace)

    for include_element in include_elements:
        element = series.find(namespace + include_element)
        if not element is None:
            name = util.camel_to_underscore(element.tag)
            element_dict = _scrub_prefix(_element_dict(element), name)
            series_dict[name] = element_dict

    return series_dict


def _parse_site(site, namespace):
    """returns a dict representation of a site given an etree object
    representing a site element
    """
    site_dict = _parse_site_info(site.find(namespace + "siteInfo"), namespace)
    series_elements = site.iter(namespace + "series")
    site_dict["series"] = [
        _parse_series(series_element, namespace) for series_element in series_elements
    ]

    return site_dict


def _parse_site_info(site_info, namespace):
    """returns a dict representation of a site given an etree object
    representing a siteInfo element
    """
    site_code = site_info.find(namespace + "siteCode")

    return_dict = {
        "code": site_code.text,
        "name": site_info.find(namespace + "siteName").text,
        "network": site_code.attrib.get("network"),
    }

    agency = site_code.attrib.get("agencyCode")
    if agency:
        return_dict["agency"] = agency

    geog_location = site_info.find(namespace.join(["", "geoLocation/", "geogLocation"]))
    if not geog_location is None:
        return_dict["location"] = _parse_geog_location(geog_location, namespace)

    timezone_info = site_info.find(namespace + "timeZoneInfo")
    if not timezone_info is None:
        return_dict["timezone_info"] = _parse_timezone_info(timezone_info, namespace)

    elevation_m = site_info.find(namespace + "elevation_m")
    if not elevation_m is None:
        return_dict["elevation_m"] = elevation_m.text

    # WaterML 1.0 notes
    notes = dict(
        [
            (util.camel_to_underscore(note.attrib["title"].replace(" ", "")), note.text)
            for note in site_info.findall(namespace + "note")
        ]
    )
    if notes:
        return_dict["notes"] = notes

    # WaterML 1.1 siteProperties
    site_properties = dict(
        [
            (
                util.camel_to_underscore(site_property.attrib["name"].replace(" ", "")),
                site_property.text,
            )
            for site_property in site_info.findall(namespace + "siteProperty")
        ]
    )
    if site_properties:
        return_dict["site_property"] = site_properties

    return return_dict


def _parse_timezone_element(timezone_element):
    """returns a dict representation of a timezone etree element (either
    defaultTimeZone or daylightSavingsTimeZone)
    """
    return {
        "abbreviation": timezone_element.attrib.get("zoneAbbreviation"),
        "offset": timezone_element.attrib.get("zoneOffset"),
    }


def _parse_timezone_info(timezone_info, namespace):
    """returns a dict representation of a timeZoneInfo etree element"""
    return_dict = {}

    uses_dst_str = timezone_info.attrib.get("siteUsesDaylightSavingsTime", "false")
    if uses_dst_str == "true":
        return_dict["uses_dst"] = True
    else:
        return_dict["uses_dst"] = False

    dst_element = timezone_info.find(namespace + "daylightSavingsTimeZone")
    if not dst_element is None:
        return_dict["dst_tz"] = _parse_timezone_element(dst_element)

    return_dict["default_tz"] = _parse_timezone_element(
        timezone_info.find(namespace + "defaultTimeZone")
    )

    return return_dict


def _parse_time_info(time_info_element, namespace):
    """returns a dict that represents a parsed WOF 1.0 timeSupport or WOF 1.1
    timeScale element
    """
    return_dict = {}

    is_regular = time_info_element.attrib.get("isRegular")
    if not is_regular is None:
        if is_regular.lower() == "true":
            is_regular = True
        elif is_regular.lower() == "false":
            is_regular = False
        return_dict["is_regular"] = is_regular

    if "1.0" in namespace:
        interval_tag = "timeInterval"
    elif "1.1" in namespace:
        interval_tag = "timeSupport"

    interval_element = time_info_element.find(namespace + interval_tag)
    if not interval_element is None:
        return_dict["interval"] = interval_element.text

    unit_element = _find_unit(time_info_element, namespace)
    if not unit_element is None:
        return_dict["units"] = _parse_unit(unit_element, namespace)

    return return_dict


def _parse_unit(unit_element, namespace):
    """returns a list of dicts that represent the values for a given unit or
    units element
    """
    unit_dict = _element_dict(unit_element)
    tag_name = unit_element.tag.split("}")[-1]
    return_dict = {}

    if "1.0" in namespace:
        return_dict["name"] = unit_element.text

    keys = ["abbreviation", "code", "name", "type"]
    for key in keys:
        dict_key = tag_name + "_" + key
        if dict_key in unit_dict:
            return_dict[key] = unit_dict[dict_key]

    return return_dict


def _parse_value(value_element, namespace):
    value_dict = _element_dict(value_element, prepend_attributes=False)
    datetime = _parse_datetime(value_dict.pop("date_time"))
    value_dict["datetime"] = datetime
    return value_dict


def _parse_values(values_element, namespace):
    """returns a list of dicts that represent the values for a given etree
    values element
    """

    return [
        _parse_value(value, namespace)
        for value in values_element.findall(namespace + "value")
    ]


def _parse_variable(variable_element, namespace):
    """returns a dict that represents a variable for a given etree variable element"""
    return_dict = _element_dict(
        variable_element,
        exclude_children=[
            "options",
            "timeScale",
            "timeSupport",
            "unit",
            "units",
            "variableCode",
            "variableDescription",
            "variableName",
        ],
    )
    variable_code = variable_element.find(namespace + "variableCode")
    return_dict.update(
        {
            "code": variable_code.text,
            "id": variable_code.attrib.get("variableID"),
            "name": variable_element.find(namespace + "variableName").text,
            "vocabulary": variable_code.attrib.get("vocabulary"),
        }
    )
    network = variable_code.attrib.get("network")
    if network:
        return_dict["network"] = network

    statistic = variable_element.find(
        namespace + "options/" + namespace + "option[@name='Statistic']"
    )
    if statistic is not None:
        return_dict["statistic"] = {
            "code": statistic.attrib.get("optionCode"),
            "name": statistic.text,
        }

    if "1.0" in namespace:
        time_info_name = "timeSupport"
    elif "1.1" in namespace:
        time_info_name = "timeScale"
    time_info_element = variable_element.find(namespace + time_info_name)
    if not time_info_element is None:
        return_dict["time"] = _parse_time_info(time_info_element, namespace)

    unit_element = _find_unit(variable_element, namespace)
    if not unit_element is None:
        return_dict["units"] = _parse_unit(unit_element, namespace)

    variable_description = variable_element.find(namespace + "variableDescription")
    if not variable_description is None:
        return_dict["description"] = variable_description.text

    return return_dict


def _scrub_prefix(element_dict, prefix):
    "returns a dict with prefix scrubbed from the keys"
    return dict([(k.split(prefix + "_")[-1], v) for k, v in element_dict.items()])



"""
Access to data provided by the `Lower Colorado River Authority`_.

`Water Quality`_ web site.
.. _Lower Colorado River Authority: http://www.lcra.org
.. _Water Quality: http://waterquality.lcra.org/
"""
from __future__ import print_function

from tsgettoolbox.ulmo.lcra.waterquality.core import (
    get_historical_data,
    get_recent_data,
)

# def get_historical_data(site_code, start=None, end=None, as_dataframe=False):
# def get_recent_data(site_code, as_dataframe=False):


def ulmo_df(site_code, start_date=None, end_date=None, historical=True):
    """Call to ulmo.lcra.waterquality.core functions.

    Specifically: .get_historical_data or .get_recent_data.
    """
    if historical is True:
        return get_historical_data(
            site_code, start=start_date, end=end_date, as_dataframe=True
        )
    return get_recent_data(site_code, as_dataframe=True)


if __name__ == "__main__":
    #    import time
    #
    #    r = ulmo_df('blah',
    #                'upperbasin')
    #
    #    print('UB EVERYTHING')
    #    print(r)
    #
    #    time.sleep(20)

    R = ulmo_df(6977, historical=True, start_date="2015-11-04", end_date="2015-12-05")

    print("UB EVERYTHING")
    print(R)



"""
    ulmo.usace.rivergages.core
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides access to data provided by the `United States Army
    Corps of Engineers`_ `Rivergages`_ web site.

    .. _United States Army Corps of Engineers: http://www.usace.army.mil/
    .. _Rivergages: http://rivergages.mvr.usace.army.mil/WaterControl/new/layout.cfm
"""
import datetime
import os.path

import requests
from bs4 import BeautifulSoup

from tsgettoolbox.ulmo import util

USACE_RIVERGAGES_DIR = os.path.join(util.get_ulmo_dir(), "usace/rivergages/")
URL = "http://rivergages.mvr.usace.army.mil/WaterControl/datamining2.cfm"
DEFAULT_START_DATE = datetime.date(1800, 1, 1)


def get_stations():
    path = os.path.join(USACE_RIVERGAGES_DIR, "datamining_field_list.cfm")

    with util.open_file_for_url(URL, path, use_bytes=True) as f:
        soup = BeautifulSoup(f, "lxml")
        options = soup.find("select", id="fld_station").find_all("option")
        stations = _parse_options(options)

    return stations


def get_station_data(
    station_code, parameter, start=None, end=None, min_value=None, max_value=None
):

    if min_value is None:
        min_value = -9000000
    if max_value is None:
        max_value = 9000000
    if start is None:
        start_date = DEFAULT_START_DATE
    else:
        start_date = util.convert_date(start)
    if end is None:
        end_date = datetime.date.today()
    else:
        end_date = util.convert_date(end)

    start_date_str = _format_date(start_date)
    end_date_str = _format_date(end_date)

    form_data = {
        "fld_station": station_code,
        "fld_parameter": parameter,
        "fld_from": min_value,
        "fld_to": max_value,
        "fld_fromdate": start_date_str,
        "fld_todate": end_date_str,
        "hdn_excel": "",
    }

    req = requests.post(URL, params=dict(sid=station_code), data=form_data)
    soup = BeautifulSoup(req.content, "lxml")
    data_table = soup.find("table").find_all("table")[-1]

    return dict([_parse_value(value_tr) for value_tr in data_table.find_all("tr")[2:]])


def get_station_parameters(station_code):
    req = requests.get(URL, params=dict(sid=station_code))
    soup = BeautifulSoup(req.content, "lxml")

    options = soup.find("select", id="fld_parameter").find_all()
    parameters = _parse_options(options)
    return parameters


def _format_date(date):
    return "%s/%s/%s" % (date.month, date.day, date.year)


def _parse_options(options):
    return dict(
        [
            (option.attrs.get("value"), option.text.strip())
            for option in options
            if option.attrs.get("value") != ""
        ]
    )


def _parse_value(value_tr):
    date_td, value_td = value_tr.find_all("td")

    return (util.convert_date(date_td.text), float(value_td.text))



from __future__ import print_function
from __future__ import absolute_import

import datetime
from base64 import b64decode
from gzip import GzipFile
import logging
import os
from builtins import object
from io import StringIO
from io import BytesIO

from tsgettoolbox.odo import convert
from tsgettoolbox.odo import odo
from tsgettoolbox.odo import resource

import pandas as pd

import requests

from tstoolbox import tsutils


class NDBC(object):
    def __init__(self, url, **query_params):
        self.lmap = {
            "stdmet": "h",
            "cwind": "c",
            "wlevel": "l",
            "ocean": "o",
            "adcp": "a",
            "supl": "s",
            "srad": "r",
        }

        # 'wlevel' header is manipulated and not in self.headermap
        self.headermap = {
            "stdmet": {
                "WDIR": "WDIR:degT",
                "WSPD": "WSPD:m/s",
                "GST": "GST:m/s",
                "WVHT": "WVHT:m",
                "DPD": "DPD:s",
                "APD": "APD:s",
                "MWD": "MWD:degT",
                "PRES": "PRES:hPa",
                "ATMP": "ATMP:degC",
                "WTMP": "WTMP:degC",
                "DEWP": "DEWP:degC",
                "VIS": "VIS:nautical_miles",
                "PTDY": "PTDY:hPa",
                "TIDE": "TIDE:ft",
            },
            "cwind": {
                "WDIR": "WDIR:degT",
                "WSPD": "WSPD:m/s",
                "GDR": "GDR:degT",
                "GST": "GST:m/s",
                "GTIME": "GTIME",
            },
            "ocean": {
                "DEPTH": "DEPTH:m",
                "OTMP": "OTMP:degC",
                "COND": "COND:mS/cm",
                "SAL": "SAL:PSU",
                "O2%": "O2:percent",
                "O2PPM": "O2PPM:ppm",
                "CLCON": "CLCON:ug/l",
                "TURB": "TURB:FTU",
                "PH": "PH:",
                "EH": "EH:mv",
            },
            "adcp": {
                "DEP01": "DEP01:m",
                "DIR01": "DIR01:degT",
                "SPD01": "SPD01:cm/s",
                "DEP02": "DEP02:m",
                "DIR02": "DIR02:degT",
                "SPD02": "SPD02:cm/s",
                "DEP03": "DEP03:m",
                "DIR03": "DIR03:degT",
                "SPD03": "SPD03:cm/s",
                "DEP04": "DEP04:m",
                "DIR04": "DIR04:degT",
                "SPD04": "SPD04:cm/s",
                "DEP05": "DEP05:m",
                "DIR05": "DIR05:degT",
                "SPD05": "SPD05:cm/s",
                "DEP06": "DEP06:m",
                "DIR06": "DIR06:degT",
                "SPD06": "SPD06:cm/s",
                "DEP07": "DEP07:m",
                "DIR07": "DIR07:degT",
                "SPD07": "SPD07:cm/s",
                "DEP08": "DEP08:m",
                "DIR08": "DIR08:degT",
                "SPD08": "SPD08:cm/s",
                "DEP09": "DEP09:m",
                "DIR09": "DIR09:degT",
                "SPD09": "SPD09:cm/s",
                "DEP10": "DEP10:m",
                "DIR10": "DIR10:degT",
                "SPD10": "SPD10:cm/s",
                "DEP11": "DEP11:m",
                "DIR11": "DIR11:degT",
                "SPD11": "SPD11:cm/s",
                "DEP12": "DEP12:m",
                "DIR12": "DIR12:degT",
                "SPD12": "SPD12:cm/s",
                "DEP13": "DEP13:m",
                "DIR13": "DIR13:degT",
                "SPD13": "SPD13:cm/s",
                "DEP14": "DEP14:m",
                "DIR14": "DIR14:degT",
                "SPD14": "SPD14:cm/s",
                "DEP15": "DEP15:m",
                "DIR15": "DIR15:degT",
                "SPD15": "SPD15:cm/s",
                "DEP16": "DEP16:m",
                "DIR16": "DIR16:degT",
                "SPD16": "SPD16:cm/s",
                "DEP17": "DEP17:m",
                "DIR17": "DIR17:degT",
                "SPD17": "SPD17:cm/s",
                "DEP18": "DEP18:m",
                "DIR18": "DIR18:degT",
                "SPD18": "SPD18:cm/s",
                "DEP19": "DEP19:m",
                "DIR19": "DIR19:degT",
                "SPD19": "SPD19:cm/s",
                "DEP20": "DEP20:m",
                "DIR20": "DIR20:degT",
                "SPD20": "SPD20:cm/s",
            },
            "srad": {
                "SRAD1": "SRAD1:w/m2",
                "SWRAD": "SWRAD:w/m2",
                "LWRAD": "LWRAD:w/m2",
            },
        }
        self.rename = {
            "WD": "WDIR:degT",
            "DIR": "WDIR:degT",
            "SPD": "WSPD:m/s",
            "GSP": "GST:m/s",
            "GMN": "GTIME",
            "BARO": "PRES:hPa",
            "H0": "WVHT:m",
            "DOMPD": "DPD:s",
            "AVP": "APD:s",
            "SRAD": "SWRAD",
            "SRAD2": "SWRAD",
            "LRAD": "LWRAD",
            "LRAD1": "LWRAD",
            "BAR": "PRES:hPa",
        }
        for item in self.headermap:
            self.rename.update(self.headermap[item])

        self.url = url
        self.query_params = query_params


# Function to make `resource` know about the new NOS type.
#  https://www.ndbc.noaa.gov/download_data.php?
#      filename=41013h2003.txt.gz&
#      dir=data/historical/stdmet/
#  https://www.ndbc.noaa.gov/download_data.php?
#      filename=4101322018.txt.gz&
#      dir=data/stdmet/Feb/
#  https://www.ndbc.noaa.gov/data/historical/stdmet/41012h2012.txt.gz
#  https://www.ndbc.noaa.gov/data/stdmet/Mar/sauf132018.txt.gz
@resource.register(r"https://www\.ndbc\.noaa\.gov/data/*", priority=17)
def resource_ndbc(uri, **kwargs):
    return NDBC(uri, **kwargs)


def date_parser(*x):
    x = [int(i) for i in x]
    if x[0] < 100:
        x[0] = x[0] + 1900
    return pd.datetime(*x)


mapnumtoname = {
    0: "Jan",
    1: "Feb",
    2: "Mar",
    3: "Apr",
    4: "May",
    5: "Jun",
    6: "Jul",
    7: "Aug",
    8: "Sep",
    9: "Oct",
    10: "Nov",
    11: "Dec",
}


# Function to convert from NDBC type to pd.DataFrame
@convert.register(pd.DataFrame, NDBC)
def ndbc_to_df(data, **kwargs):

    sdate = tsutils.parsedate(data.query_params.pop("startUTC"))
    edate = tsutils.parsedate(data.query_params.pop("endUTC"))

    table = data.query_params["table"]

    df = pd.DataFrame()

    cyear = datetime.datetime.now()
    filenames = []
    for yr in range(sdate.year, edate.year + 1):
        # Yearly
        # https://www.ndbc.noaa.gov/data/historical/stdmet/41012h2012.txt.gz
        filenames.append(
            "/historical/{3}/{0}{1}{2}.txt.gz".format(
                data.query_params["station"], data.lmap[table], yr, table
            )
        )
    if edate.year == cyear.year:
        for mnth in range(edate.month):
            # Monthly
            # https://www.ndbc.noaa.gov/data/stdmet/Mar/sauf132018.txt.gz
            filenames.append(
                "/{3}/{4}/{0}{1}{2}.txt.gz".format(
                    data.query_params["station"],
                    mnth + 1,
                    yr,
                    table,
                    mapnumtoname[mnth],
                )
            )

    for filename in filenames:
        req = requests.get(data.url + filename)

        if os.path.exists("debug_tsgettoolbox"):
            logging.warning(req.url)

        try:
            req.raise_for_status()
        except requests.exceptions.HTTPError:
            continue

        content = BytesIO(req.content)
        content = GzipFile(fileobj=content).read()

        # Test to see if 1 or 2 line header
        f = StringIO(content.decode("utf-8"))
        head1 = f.readline()
        head2 = f.readline()
        f.seek(0)

        words1 = head1.split()
        words2 = head2.split()
        try:
            _ = int(words2[0])
            skiprows = None
        except ValueError:
            skiprows = [1]

        if "mm" == words1[4]:
            parse_dates = {"datetime": [0, 1, 2, 3, 4]}
        else:
            parse_dates = {"datetime": [0, 1, 2, 3]}

        tdf = pd.read_csv(
            f,
            header=0,
            skiprows=skiprows,
            sep=r"\s+",
            parse_dates=parse_dates,
            date_parser=date_parser,
            index_col=0,
            na_values=["MM", 999.0, 99.0, 9999, 99999],
        )

        if len(tdf) > 0:
            tdf.rename(columns=data.rename, inplace=True)
            df = df.append(tdf)

    if len(df) == 0:
        raise ValueError(
            tsutils.error_wrapper(
                """
No data collected/available within this time frame.
"""
            )
        )

    if table == "wlevel":
        df.columns = list(range(0, 55, 6))
        df = pd.DataFrame(df.stack())
        df.index = [i + datetime.timedelta(minutes=j) for i, j in df.index]
        df.index.name = "Datetime"
        df.columns = ["WLEVEL:ft:MLLW"]

    # Clean up the dataframe...
    df = df.sort_index()
    df = df[~df.index.duplicated()]

    df.columns = [i.replace(r"%", "PERCENT") for i in df.columns]

    return df.loc[sdate:edate, :]


if __name__ == "__main__":
    """
    """

    r = resource(
        r"https://www.ndbc.noaa.gov/data/",
        table="stdmet",
        startUTC="2012-01-01T00:00Z",
        endUTC="2012-04-01T00:00Z",
        station="41012",
    )

    as_df = odo(r, pd.DataFrame)
    print("NDBC")
    print(as_df)
    as_df.to_csv("file.csv")



from __future__ import absolute_import, division, print_function

import pytest

bokeh = pytest.importorskip("bokeh")

from tsgettoolbox.odo.backends.bokeh import convert, pd, ColumnDataSource
import pandas.util.testing as tm


df = pd.DataFrame(
    [[100, "Alice"], [200, "Bob"], [300, "Charlie"]], columns=["balance", "name"]
)


def test_convert_dataframe_to_cds():
    cds = convert(ColumnDataSource, df)
    assert list(cds.data["name"]) == ["Alice", "Bob", "Charlie"]
    assert list(cds.data["balance"]) == [100, 200, 300]
    df2 = convert(pd.DataFrame, cds)
    assert isinstance(df2, pd.DataFrame)

    tm.assert_frame_equal(df, df2)



"""
    ulmo.cuahsi.wof
    ~~~~~~~~~~~~~~~

    `CUAHSI WaterOneFlow`_ web services

    .. _CUAHSI WaterOneFlow: http://his.cuahsi.org/wofws.html
"""
from __future__ import absolute_import

from . import core

from .core import get_sites, get_site_info, get_values, get_variable_info



from __future__ import absolute_import, division, print_function

from tsgettoolbox.odo.backends.text import (
    TextFile,
    resource,
    convert,
    discover,
    append,
    drop,
    chunks,
    Temp,
)
from tsgettoolbox.odo.utils import tmpfile, filetexts, filetext
from datashape import dshape
import gzip
import os


def test_resource():
    assert isinstance(resource("foo.txt"), TextFile)
    assert isinstance(resource("foo.log.gz"), TextFile)
    assert isinstance(resource("/path/to/foo.log"), TextFile)


def test_open():
    r = resource("foo.log.gz")
    assert r.open == gzip.open


def test_convert():
    with filetext("Hello\nWorld") as fn:
        assert convert(list, TextFile(fn)) == ["Hello\n", "World"]


def test_append():
    with tmpfile("log") as fn:
        t = TextFile(fn)
        append(t, ["Hello", "World"])

        assert os.path.exists(fn)
        with open(fn) as f:
            assert list(map(str.strip, f.readlines())) == ["Hello", "World"]


def test_discover():
    assert discover(TextFile("")) == dshape("var * string")


def test_drop():
    with filetext("hello\nworld") as fn:
        t = TextFile(fn)
        assert os.path.exists(fn)
        drop(t)
        assert not os.path.exists(fn)


def test_chunks_textfile():
    with filetexts({"a1.log": "Hello\nWorld", "a2.log": "Hola\nMundo"}) as fns:
        logs = chunks(TextFile)(list(map(TextFile, fns)))
        assert set(map(str.strip, convert(list, logs))) == set(
            ["Hello", "World", "Hola", "Mundo"]
        )


def test_temp():
    t = convert(Temp(TextFile), [1, 2, 3])
    assert [int(line.strip()) for line in convert(list, t)] == [1, 2, 3]



from __future__ import absolute_import, division, print_function

import pytest

pymongo = pytest.importorskip("pymongo")

import os
from tsgettoolbox.odo import discover, convert, append, resource, dshape, odo
from toolz import pluck
from copy import deepcopy
from bson.objectid import ObjectId


@pytest.fixture(scope="module")
def mongo_host_port():
    import os

    return (
        os.environ.get("MONGO_IP", "localhost"),
        os.environ.get("MONGO_PORT", 27017),
    )


@pytest.fixture(scope="module")
def conn(mongo_host_port):
    host, port = mongo_host_port
    try:
        return pymongo.MongoClient(host=host, port=port)
    except pymongo.errors.ConnectionFailure:
        pytest.skip("No mongo server running")


@pytest.yield_fixture
def db(conn):
    try:
        yield conn._test_db
    finally:
        conn.drop_database("_test_db")


@pytest.fixture
def raw_bank():
    return [
        {"name": "Alice", "amount": 100},
        {"name": "Alice", "amount": 200},
        {"name": "Bob", "amount": 100},
        {"name": "Bob", "amount": 200},
        {"name": "Bob", "amount": 300},
    ]


@pytest.yield_fixture
def bank(db, raw_bank):
    db.bank.insert(deepcopy(raw_bank))
    try:
        yield db.bank
    finally:
        db.drop_collection("bank")


@pytest.yield_fixture
def empty_bank(db):
    try:
        yield db.empty_bank
    finally:
        db.drop_collection("empty_bank")


ds = dshape("var * {name: string, amount: int}")


def test_discover(bank, raw_bank):
    assert discover(bank) == discover(raw_bank)


def test_discover_empty_db(db):
    # NOTE: nothing has been added to the database because our fixtures create
    # a new db and collections for each test that depends on them, so
    # 'system.indexes' won't exist here yet
    # The same is true in test_resource_db
    assert discover(db).measure.names == []


def test_discover_db(bank, db):
    assert "bank" in set(discover(db).measure.names)


def test_resource_db(mongo_host_port):
    db = resource("mongodb://{}:{}/_test_db".format(*mongo_host_port))
    assert db.name == "_test_db"
    assert discover(db).measure.names == []


def test_resource_collection(mongo_host_port):
    host, port = mongo_host_port
    coll = resource("mongodb://{}:{}/db::mycoll".format(*mongo_host_port))
    assert coll.name == "mycoll"
    assert coll.database.name == "db"
    assert coll.database.connection.host == host
    assert coll.database.connection.port == port


def test_append_convert(empty_bank, raw_bank):
    ds = discover(raw_bank)
    assert set(ds.measure.names) == {"name", "amount"}

    append(empty_bank, raw_bank, dshape=ds)
    assert odo(empty_bank, list, dshape=ds) == list(pluck(ds.measure.names, raw_bank))


@pytest.yield_fixture
def multiple_object_ids(db):
    data = [
        {"x": 1, "y": 2, "other": ObjectId("1" * 24)},
        {"x": 3, "y": 4, "other": ObjectId("2" * 24)},
    ]
    db.multiple_object_ids.insert(data)
    try:
        yield db.multiple_object_ids
    finally:
        db.drop_collection("multiple_object_ids")


def test_multiple_object_ids(multiple_object_ids):
    assert discover(multiple_object_ids) == dshape("2 * {x: int64, y: int64}")
    assert convert(list, multiple_object_ids) == [(1, 2), (3, 4)]



import pytest

pytest.importorskip("dask.bag")

from operator import methodcaller

from tsgettoolbox.odo import chunks, TextFile, odo
from dask.bag import Bag
from tsgettoolbox.odo.utils import filetexts


def inc(x):
    return x + 1


dsk = {("x", 0): (range, 5), ("x", 1): (range, 5), ("x", 2): (range, 5)}

L = list(range(5)) * 3

b = Bag(dsk, "x", 3)


def test_convert_bag_to_list():
    assert odo(b, list) == L


def test_convert_logfiles_to_bag():
    with filetexts({"a1.log": "Hello\nWorld", "a2.log": "Hola\nMundo"}) as fns:
        logs = chunks(TextFile)(list(map(TextFile, fns)))
        b = odo(logs, Bag)
        assert isinstance(b, Bag)
        assert list(map(methodcaller("strip"), odo(b, list))) == list(
            map(methodcaller("strip"), odo(logs, list))
        )


def test_sequence():
    b = odo([1, 2, 3], Bag)
    assert set(b.map(inc)) == set([2, 3, 4])



from __future__ import absolute_import, division, print_function

import os


def test_create_pydot_image():
    try:
        from tsgettoolbox.odo.dot import dot_graph

        dot_graph()
    except:
        pass
    finally:
        for ext in ("pdf", "png"):
            try:
                os.remove("conversions.%s" % ext)
            except:
                pass



"""
    ulmo.cuahsi.his_central
    ~~~~~~~~~~~~~~~~~~~~~~~

    `CUAHSI HIS Central`_ web services

    .. _CUAHSI HIS Central: http://his.cuahsi.org/hiscentral.html
"""
from .core import get_services



from __future__ import absolute_import, division, print_function

import pytest

from tsgettoolbox.odo.numpy_dtype import (
    dshape_to_pandas,
    unit_to_dtype,
    dshape_to_numpy,
)
from datashape import dshape
import numpy as np


@pytest.mark.parametrize(
    ["ds", "expected"],
    [
        ("decimal[9,2]", np.float64),
        ("decimal[9]", np.int32),
        ("?decimal[9]", np.float32),
        ("?decimal[1,0]", np.float16),
    ],
)
def test_decimal(ds, expected):
    assert unit_to_dtype(dshape(ds)) == expected


@pytest.mark.parametrize(
    ["ds", "field"],
    [
        ("var * {teststr1: option[string[4]]}", "teststr1"),
        ('var * {teststr2: option[string["ascii"]]}', "teststr2"),
    ],
)
def test_parameterized_option_instances(ds, field):
    dtypes, _ = dshape_to_pandas(dshape(ds))
    assert isinstance(dtypes[field], np.dtype)


@pytest.mark.parametrize(
    "ds", ['option[datetime[tz="EST"]]', 'option[timedelta[unit="D"]]']
)
def test_unit_to_dtype(ds):
    assert isinstance(unit_to_dtype(ds), np.dtype)


@pytest.mark.parametrize(
    ["ds", "expected"],
    [
        ("{a: int32}", ({"a": np.dtype("int32")}, [])),
        ("{a: int32, when: datetime}", ({"a": np.dtype("int32")}, ["when"])),
        ("{a: ?int64}", ({"a": np.dtype("float64")}, [])),
    ],
)
def test_dshape_to_pandas(ds, expected):
    assert dshape_to_pandas(ds) == expected


@pytest.mark.parametrize(
    ["ds", "dt"],
    [
        ("int32", "int32"),
        ("?int32", "float32"),
        (
            '{name: string[5, "ascii"], amount: ?int32}',
            [("name", "S5"), ("amount", "<f4")],
        ),
        ("(int32, float32)", [("f0", "<i4"), ("f1", "<f4")]),
    ],
)
def test_dshape_to_numpy(ds, dt):
    assert dshape_to_numpy(ds) == np.dtype(dt)



from __future__ import print_function

import pytest
from mock import patch, Mock

from functools import partial
import codecs
import os

from tsgettoolbox.odo import odo, resource, URL, discover, CSV, TextFile, convert
from tsgettoolbox.odo.backends.url import sample
from tsgettoolbox.odo.temp import _Temp, Temp
from tsgettoolbox.odo.utils import tmpfile, raises

import datashape

try:
    from urllib2 import urlopen
    from urllib2 import HTTPError, URLError
except ImportError:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError

pytestmark = pytest.mark.skipif(
    raises(URLError, partial(urlopen, "http://google.com")),
    reason="unable to connect to google.com",
)

iris_url = (
    "https://raw.githubusercontent.com/"
    "blaze/blaze/master/blaze/examples/data/iris.csv"
)
ftp_url = "ftp://athena-dist.mit.edu/pub/XNeXT/README.txt"


def test_url_resource():
    csv = resource(iris_url)
    assert isinstance(csv, URL(CSV))


def test_small_chunk_size():
    normal = convert(Temp(CSV), resource(iris_url))
    small_chunk = convert(Temp(CSV), resource(iris_url, chunk_size=1))
    with open(normal.path, "rb") as fn:
        normal_data = fn.read()
    with open(small_chunk.path, "rb") as fn:
        small_chunk_data = fn.read()
    assert normal_data == small_chunk_data


def test_sample_different_line_counts():
    with sample(resource(iris_url), lines=10) as fn:
        with open(fn, "r") as f:
            assert len(list(f)) == 10

    with sample(resource(iris_url), lines=5) as fn:
        with open(fn, "r") as f:
            assert len(list(f)) == 5


def test_sample_different_encoding():
    encoding = "latin-1"
    lines = 10
    with sample(resource(iris_url), lines=lines, encoding=encoding) as fn:
        with codecs.open(fn, "r", encoding=encoding) as f:
            assert len(list(f)) == lines


@pytest.mark.xfail(raises=HTTPError)
@patch("tsgettoolbox.odo.backends.url.urlopen")
def test_failed_url(m):
    failed_url = "http://foo.com/myfile.csv"
    m.side_effect = HTTPError(failed_url, 404, "Not found", None, None)
    with tmpfile(".csv") as fn:
        odo(failed_url, fn)


def test_url_discover():
    csv = resource(iris_url)
    assert isinstance(discover(csv), datashape.DataShape)


def test_url_to_local_csv():
    with tmpfile(".csv") as fn:
        csv = odo(iris_url, fn)
        path = os.path.abspath(csv.path)
        assert os.path.exists(path)


def test_url_txt_resource():
    txt = resource(ftp_url)
    assert isinstance(txt, URL(TextFile))


@pytest.mark.xfail(
    raises=URLError, reason="MIT Athena FTP is down as of October 23, 2015"
)
def test_ftp_to_local_txt():
    with tmpfile(".txt") as fn:
        txt = odo(ftp_url, fn, timeout=5)
        path = os.path.abspath(txt.path)
        assert os.path.exists(path)


def test_convert():
    url_csv = resource(iris_url)
    t_csv = convert(Temp(CSV), url_csv)
    assert discover(url_csv) == discover(t_csv)

    assert isinstance(t_csv, _Temp)


@pytest.mark.skipif(
    os.environ.get("HDFS_TEST_HOST") is None, reason="No HDFS_TEST_HOST envar defined"
)
def test_url_to_hdfs():
    from .test_hdfs import tmpfile_hdfs, hdfs, HDFS

    with tmpfile_hdfs() as target:

        # build temp csv for assertion check
        url_csv = resource(iris_url)
        csv = convert(Temp(CSV), url_csv)

        # test against url
        scsv = HDFS(CSV)(target, hdfs=hdfs)
        odo(iris_url, scsv)

        assert discover(scsv) == discover(csv)



"""
This module provides direct access to `Climate Predicition Center`_.

`Weekly Drought Index`_ dataset from `Climate Prediction Center`_.

.. _Climate Prediction Center: http://www.cpc.ncep.noaa.gov/
.. _Weekly Drought Index: http://www.cpc.ncep.noaa.gov/products/analysis_monitoring/cdus/palmer_drought/
"""
from __future__ import print_function
from __future__ import absolute_import

import pandas as pd

from tsgettoolbox.ulmo.cpc.drought.core import get_data


def ulmo_df(state=None, climate_division=None, start_date=None, end_date=None):
    df = get_data(
        state=state,
        climate_division=climate_division,
        start=start_date,
        end=end_date,
        as_dataframe=True,
    )
    df = df.set_index("period")
    df.index = pd.PeriodIndex(df.index)
    df.index.name = "Datetime"
    return df


if __name__ == "__main__":
    r = ulmo_df(
        state="FL", climate_division=1, start_date="2017-01-01", end_date="2017-10-02"
    )

    print("FL EVERYTHING")
    print(r)
